from PySide2 import QtCore
from PySide2 import QtWidgets
from PySide2 import QtGui
from PySide2.QtCore import *

from piscat.InputOutput import write_video
from piscat.Preproccessing.normalization import Normalization
from piscat.Visualization.display import Display, DisplayPSFs_subplotLocalizationDisplay
from piscat.Visualization.display_html import HTML_PSFs_subplotLocalizationDisplay

from piscat.GUI.Visualization.contrast_adjustment_GUI import Contrast_adjustment_GUI
from piscat.GUI.Visualization import slice_view
from piscat.GUI.InputOutput import save_GUI
from piscat.GUI.Visualization.play_setting import PlaySetting

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import os

play_flag = True


class WorkerSignals(QObject):

    updateProgress = Signal(int)
    result = Signal(object)
    finished = Signal()
    stop_signal = Signal()


class ImageViewer(QtWidgets.QMainWindow, QtCore.QObject):

    progressChanged = QtCore.Signal(int)
    sliceNumber = QtCore.Signal(int)
    currentFrame = QtCore.Signal(int)

    def __init__(self, input_image_series, original_video, name, mask=False, list_psf=None):
        super(ImageViewer, self).__init__()

        self.threadpool = QThreadPool()

        self.image_series = input_image_series
        self.original_video = original_video
        self.df_PSFs = list_psf
        if self.image_series.ndim == 3:
            self.sizeX = self.image_series.shape[1]
            self.sizeY = self.image_series.shape[2]
        elif self.image_series.ndim == 2:
            self.sizeX = self.image_series.shape[0]
            self.sizeY = self.image_series.shape[1]
        self.frame_num = None
        self.current_frame = 0
        self.title = name
        self.mask = mask
        self.starter()
        self.save_video = None
        self.step_size = None
        self.time_delay = None
        self.fps = None
        self.info_image_save = None
        self.alpha = None
        self.beta = None
        self.max_intensity = None
        self.min_intensity = None
        self.frame_strides = 1
        self.signals = WorkerSignals()
        self.sliceNumber.connect(self.Update_SliceNumber)

    def __del__(self):
        print('Destructor called, Employee deleted.')

    def starter(self):
        self.setWindowIcon(QtGui.QIcon(QtCore.QDir.currentPath() + "/icons/mpl.png"))
        self.setWindowTitle(self.title)
        self.setGeometry(490, 100,  512+124,  512+93)
        self.main_Viewer_frame = QtWidgets.QFrame()
        self.setCentralWidget(self.main_Viewer_frame)
        self.main_Viewer_frame_layout = QtWidgets.QGridLayout()
        self.main_Viewer_frame.setLayout(self.main_Viewer_frame_layout)
        self.resize(500, 500)

        # Create an instant of GraphicView Class to show the image PixMap.
        self.viewer = slice_view.SliceView(self.image_series, self.original_video, mask=self.mask)
        self.viewer.frameClicked.connect(self.frame_read)
        self.viewer.photoClicked.connect(self.photoClicked)

        self.main_Viewer_frame_layout.addWidget(self.viewer, 0, 0)

        # checkbox median Filter
        self.checkbox_medianFilter = QtWidgets.QRadioButton("Median Filter")
        self.checkbox_medianFilter.toggled.connect(lambda: self.activeMedianFilter(self.checkbox_medianFilter))

        # Contrast
        self.con_adj = QtWidgets.QPushButton("Contrast adjustment")
        self.con_adj.setAutoDefault(False)
        self.con_adj.clicked.connect(self.open_adjustment)
        self.con_adj.setFixedWidth(130)
        self.con_adj.setStyleSheet("background-color : lightgrey")

        # Display
        self.Display = QtWidgets.QPushButton("Display")
        self.Display.setAutoDefault(False)
        self.Display.clicked.connect(self.matplotlib_display)
        self.Display.setFixedWidth(130)
        self.Display.setStyleSheet("background-color : lightgrey")

        # annotation
        self.annotation = QtWidgets.QPushButton("Live annotation")
        self.annotation.setCheckable(True)
        self.annotation.clicked.connect(self.annotationFun)
        self.annotation.setFixedWidth(130)
        self.annotation.setStyleSheet("background-color : lightgrey")

        # Play
        self.playBtn = QtWidgets.QPushButton()
        self.playBtn.setAutoDefault(False)
        self.playBtn.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPlay))
        self.playBtn.setFixedWidth(20)
        self.playBtn.clicked.connect(self.run_play)

        # create context menu
        self.playBtn.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.playBtn.customContextMenuRequested.connect(self.on_context_menu)

        # stop
        self.stopBtn = QtWidgets.QPushButton()
        self.stopBtn.setAutoDefault(False)
        self.stopBtn.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaStop))
        self.stopBtn.setFixedWidth(20)
        self.stopBtn.clicked.connect(self.stop_video)

        # Pause
        self.pauseBtn = QtWidgets.QPushButton()
        self.pauseBtn.setAutoDefault(False)
        self.pauseBtn.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPause))
        self.pauseBtn.setFixedWidth(20)
        self.pauseBtn.clicked.connect(self.pause_video)
        self._running = False

        # save
        self.save = QtWidgets.QPushButton("save")
        self.save.setAutoDefault(False)
        self.save.clicked.connect(self.do_update_save)
        self.save.setFixedWidth(130)
        self.save.setStyleSheet("background-color : lightgrey")

        if self.mask is True:
            self.Display_mask = QtWidgets.QPushButton("Display_with_mask")
            self.Display_mask.setAutoDefault(False)
            self.Display_mask.clicked.connect(self.matplotlib_display_mask)
            self.Display_mask.setFixedWidth(130)
            self.Display_mask.setStyleSheet("background-color : lightgrey")

            # mp4
            self.mp4 = QtWidgets.QPushButton("MP4+localization")
            self.mp4.setAutoDefault(False)
            self.mp4.clicked.connect(self.make_mp4)
            self.mp4.setFixedWidth(130)
            self.mp4.setStyleSheet("background-color : lightgrey")

            # progressBar
            self.statusbar = self.statusBar()
            self.progressBar = QtWidgets.QProgressBar(self)
            self.statusBar().addPermanentWidget(self.progressBar)
            self.progressBar.setGeometry(30, 40, 200, 25)
            self.progressBar.setValue(0)
            self.progressChanged.connect(self.setProgress)

        # Create
        self.connect(QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_H), self), QtCore.SIGNAL('activated()'), self.histogram)

        if self.title != "PNG":
            self.slice_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
            self.slice_slider.setMinimum(0)
            self.slice_slider.setMaximum(self.image_series.shape[0]-1)
            self.slice_slider.setValue(0)
            self.slice_slider.valueChanged.connect(self.set_new_slice_num)

            self.main_Viewer_frame_layout.addWidget(self.createThirdExclusiveGroup(), 1, 0)

        self.main_Viewer_frame_layout.addWidget(self.createSecondExclusiveGroup(), 2, 0)

        self.statusbar = self.statusBar()
        self.show()

    def createSecondExclusiveGroup(self):

        self.groupBox_displayBtn = QtWidgets.QGroupBox("Video Display")

        self.grid_diplay = QtWidgets.QGridLayout()
        self.grid_diplay.addWidget(self.Display, 0, 0)
        self.grid_diplay.addWidget(self.annotation, 0, 1)
        self.grid_diplay.addWidget(self.con_adj,  0, 2)

        if self.mask is True:
            self.grid_diplay.addWidget(self.Display_mask, 1, 0)
            # self.grid_diplay.addWidget(self.mp4, 1, 1)

        self.grid_diplay.addWidget(self.save, 1, 2)
        self.groupBox_displayBtn.setLayout(self.grid_diplay)

        return self.groupBox_displayBtn

    def createThirdExclusiveGroup(self):

        play = QtWidgets.QGroupBox()

        self.grid_play = QtWidgets.QGridLayout()
        self.grid_play.addWidget(self.playBtn, 0, 0)
        self.grid_play.addWidget(self.pauseBtn, 0, 1)
        self.grid_play.addWidget(self.stopBtn, 0, 2)
        self.grid_play.addWidget(self.slice_slider, 0, 3)
        self.grid_play.addWidget(self.checkbox_medianFilter, 0, 4)

        play.setLayout(self.grid_play)
        return play

    def activeMedianFilter(self, box_status):
        if box_status.isChecked():
            self.viewer.medianFilterFlag = True
        else:
            self.viewer.medianFilterFlag = False

    def annotationFun(self):
        if self.annotation.isChecked():
            self.annotation.setStyleSheet("background-color : lightblue")

        else:
            self.annotation.setStyleSheet("background-color : lightgrey")

        self.viewer.annotation.emit()

    def closeEvent(self, event):
        try:
            self.stop_video()
        except:
            pass
        print("closed")

    @QtCore.Slot()
    def breakAll(self):
        self.make_mp4.kill()

    @QtCore.Slot()
    def Update_SliceNumber(self, frame_number):
        self.current_frame = frame_number
        self.currentFrame.emit(frame_number)

    def get_in(self, df_PSFs):
        self.df_PSFs = df_PSFs

    @QtCore.Slot()
    def call_save(self):
        if self.fps == '' or self.fps is None:
            self.fps = 10
        else:
            self.fps = int(self.fps)

        if self.info_image_save.video_type == "RAW":
            if self.info_image_save.bin_type is not None:
                type = self.info_image_save.bin_type
            else:
                type = 'original'

            file_name = str(self.original_video.shape[0]) + '_' + str(self.original_video.shape[1]) +'_' + \
                        str(self.original_video.shape[2]) + '_' + str(type) + '.raw'
            write_video.write_binary(dir_path=self.file_path, file_name=file_name, data=self.original_video, type=type)

        elif self.info_image_save.video_type == "MP4":

            self.flag_update_normalization = True
            worker = Normalization(video=self.original_video, flag_image_specific=True)
            worker.signals.result.connect(self.update_normalization)
            self.threadpool.start(worker)
            while self.flag_update_normalization:
                QtCore.QCoreApplication.processEvents()
            file_name = str(self.original_video.shape[0]) + '_' + str(self.original_video.shape[1]) + '_' + \
                        str(self.original_video.shape[2]) + '_' + str(self.original_video.dtype) + '.mp4'
            write_video.write_MP4(dir_path=self.file_path, file_name=file_name, data=self.save_video, jump=self.frame_strides, fps=self.fps)

        elif self.info_image_save.video_type == "GIF":
            self.flag_update_normalization = True
            worker = Normalization(video=self.original_video, flag_image_specific=True)
            worker.signals.result.connect(self.update_normalization)
            self.threadpool.start(worker)
            while self.flag_update_normalization:
                QtCore.QCoreApplication.processEvents()
            file_name = str(self.original_video.shape[0]) + '_' + str(self.original_video.shape[1]) + '_' + \
                        str(self.original_video.shape[2]) + '_' + str(self.original_video.dtype) + '.gif'

            write_video.write_GIF(dir_path=self.file_path, file_name=file_name, data=self.save_video, jump=self.frame_strides, fps=self.fps)

    @Slot()
    def update_normalization(self, video):
        self.save_video = video
        self.flag_update_normalization = False

    def do_update_save(self):
        if self.save.clicked:
            try:
                self.info_image_save = save_GUI.SaveVideo()
                self.info_image_save.window.show()

                self.info_image_save.signal_save_Done.connect(self.call_save)
                self.file_path = self.askdir()

            except:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Folder does not defined!")
                self.msg_box.exec_()

    def askdir(self):
        folder = QtWidgets.QFileDialog.getExistingDirectory(None, 'Select a folder:', '',
                                                                 QtWidgets.QFileDialog.ShowDirsOnly)
        return folder

    def matplotlib_display(self):

        if self.time_delay == '' or self.time_delay is None:
            self.time_delay = 0.1
        else:
            self.time_delay = float(self.time_delay)

        if self.Display.clicked:
            if self.step_size != '' and self.step_size is not None:
                Display(video=np.fliplr(self.original_video), time_delay=float(self.time_delay))
            else:
                Display(video=np.fliplr(self.original_video), time_delay=float(self.time_delay))

    def matplotlib_display_mask(self):

        if self.step_size == '' or self.step_size is None:
            self.step_size = 1
        else:
            self.step_size = int(self.step_size)

        if self.time_delay == '' or self.time_delay is None:
            self.time_delay = 0.1
        else:
            self.time_delay = float(self.time_delay)

        if self.Display_mask.clicked:
            if self.df_PSFs is not None:
                if type(self.df_PSFs) is pd.core.frame.DataFrame:
                    if self.step_size != '':

                        ani_ = DisplayPSFs_subplotLocalizationDisplay(list_videos=[self.original_video],
                                                                                list_df_PSFs=[self.df_PSFs], list_titles=None,
                                                                                numRows=1, numColumns=1, color='gray',
                                                                                median_filter_flag=self.viewer.medianFilterFlag,
                                                                                imgSizex=10, imgSizey=5, time_delay=self.time_delay)

                    else:
                        ani_ = DisplayPSFs_subplotLocalizationDisplay(list_videos=[self.original_video],
                                                                                list_df_PSFs=[self.df_PSFs], list_titles=None,
                                                                                numRows=1, numColumns=1, color='gray',
                                                                                median_filter_flag=self.viewer.medianFilterFlag,
                                                                                imgSizex=10, imgSizey=5, time_delay=self.time_delay)

            else:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("The PSF location is not defined")
                self.msg_box.exec_()

    def make_mp4(self):

        if self.mp4.clicked:
            self.file_path = self.askdir()
            try:
                if self.step_size == '' or self.step_size is None:
                    self.step_size = 1
                else:
                    self.step_size = int(self.step_size)

                if self.time_delay == '' or self.time_delay is None:
                    self.time_delay = 0.1
                else:
                    self.time_delay = float(self.time_delay)

                if self.fps == '' or self.fps is None:
                    self.fps = 10
                else:
                    self.fps = int(self.fps)

                p_max = self.original_video.shape[0] - self.step_size
                self.progressBar.setValue(0)
                self.progressBar.setMaximum(p_max)

                if self.df_PSFs is not None:
                    if type(self.df_PSFs) is pd.core.frame.DataFrame:
                        self.flag_mp4 = True
                        save_path = os.path.join(self.file_path, 'PiSCAT_GUI_Localization.mp4')

                        ani_ = HTML_PSFs_subplotLocalizationDisplay(list_videos=[self.original_video],
                                                                          list_df_PSFs=[self.df_PSFs], list_titles=None,
                                                                          numRows=1, numColumns=1, cmap='gray',
                                                                          median_filter_flag=self.viewer.medianFilterFlag,
                                                                          imgSizex=10, imgSizey=5, time_delay=self.time_delay,
                                                                          save_path=save_path, fps=self.fps)
                        ani_.signals.finished.connect(self.finish_mp4)
                        self.threadpool.start(ani_)
                        while self.flag_mp4:
                            QtCore.QCoreApplication.processEvents()

                        self.msg_box = QtWidgets.QMessageBox()
                        self.msg_box.setWindowTitle("Warning!")
                        self.msg_box.setText("MP4 vide saved!")
                        self.msg_box.exec_()

                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("The PSF location is not defined")
                    self.msg_box.exec_()
            except:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Error occurred in MP4!")
                self.msg_box.exec_()

    @QtCore.Slot()
    def finish_mp4(self):
        self.flag_mp4 = False

    def startProgressBar(self, fun, **kwargs):
        self.thread = fun(**kwargs)
        self.thread.obj_connection.updateProgress_.connect(self.setProgress)
        self.thread.obj_connection.finished.connect(self.thread_complete)
        self.threadpool.start(self.thread)

    @QtCore.Slot()
    def update_animate(self, setting):
        self.step_size = setting[0]
        self.time_delay = setting[1]
        self.fps = setting[2]
        self.raise_()

    def set_new_slice_num(self):
        self.viewer.slice_num = self.slice_slider.value()
        self.updata_viewer()
        self.sliceNumber.emit(self.viewer.slice_num)

    def enterEvent(self, event):
        self.setMouseTracking(True)

    def leaveEvent(self, event):
        self.setMouseTracking(False)

    def update_potion(self, result_):
        self.statusbar.showMessage(result_)

    def thread_complete(self):
        print("THREAD COMPLETE!")

    def wheelEvent(self, *args, **kwargs):
        pass

    def histogram(self):
        tmp = self.original_video[self.viewer.slice_num]
        plt.figure(figsize=(6, 6), dpi=100)
        plt.hist(tmp.ravel(), fc='g', ec='k')
        plt.title("Histogram of pixel intensity")
        plt.ylabel("#Counts")
        plt.show()

    @QtCore.Slot()
    def run_play(self):

        if self.step_size == '' or self.step_size is None:
            self.frame_strides = 1
        else:
            self.frame_strides = int(self.step_size)

        if self.time_delay == '' or self.time_delay is None:
            self.time_delay = 0.1
        else:
            self.time_delay = float(self.time_delay)

        self.playBtn.setDisabled(True)
        self.play_video()

    @QtCore.Slot()
    def play_video(self):
        global play_flag
        play_flag = True
        slice_num = self.slice_slider.value()
        self.playBtn.setDisabled(True)

        while slice_num < self.image_series.shape[0]-1 and play_flag:
            slice_num += self.frame_strides
            self.slice_slider.setValue(slice_num)
            self.viewer.slice_num = self.slice_slider.value()
            self.updata_viewer()
            QtCore.QCoreApplication.processEvents()

    @QtCore.Slot()
    def pause_video(self):
        global play_flag
        play_flag = False
        self.playBtn.setEnabled(True)

    @QtCore.Slot()
    def stop_video(self):

        global play_flag
        play_flag = False

        self.playBtn.setEnabled(True)

        self.slice_slider.setValue(0)
        self.viewer.slice_num = self.slice_slider.value()
        self.updata_viewer()

    def on_context_menu(self):
        self.animatePanel_ = PlaySetting()
        self.animatePanel_.procDone_Animation_panel.connect(self.update_animate)
        self.animatePanel_.show()

    def frame_read(self, frame_num):
        if self.viewer.dragMode() == QtWidgets.QGraphicsView.NoDrag:
            self.frame_num = frame_num

    def photoClicked(self, pos):
        if self.viewer.dragMode() == QtWidgets.QGraphicsView.NoDrag:
            value = self.original_video[self.frame_num, pos[0], pos[1]]
            self.statusbar.showMessage('X:%d, Y:%d, Frame:%d, Value:%f' % (pos[0], pos[1], self.frame_num, value))

    @QtCore.Slot(int)
    def setProgress(self, val):
        self.progressBar.setValue(val)

    @QtCore.Slot()
    def set_alpha(self, alpha):
        self.alpha = alpha

    @QtCore.Slot()
    def set_beta(self, beta):
        self.beta = beta

    @QtCore.Slot()
    def set_max_intensity(self, max_intensity):
        self.max_intensity = max_intensity
        self.updata_viewer()

    @QtCore.Slot()
    def set_min_intensity(self, min_intensity):
        self.min_intensity = min_intensity

    def open_adjustment(self):
        img_ = self.viewer.input_file
        min_intensity = img_.min()
        max_intensity = img_.max()
        alpha = 255 / (max_intensity - min_intensity)
        beta = -min_intensity * alpha
        self.con_adj_setting = Contrast_adjustment_GUI(alpha, beta, min_intensity, max_intensity)

        self.con_adj_setting.alpha_signal.connect(self.set_alpha)
        self.con_adj_setting.beta_signal.connect(self.set_beta)
        self.con_adj_setting.min_intensity_signal.connect(self.set_min_intensity)
        self.con_adj_setting.max_intensity_signal.connect(self.set_max_intensity)

    def updata_viewer(self):
        self.viewer.current_pixmap = self.viewer.create_pixmap(self.image_series[self.viewer.slice_num])
        self.viewer.update_slice(self.viewer.current_pixmap)
        if self.viewer.mask_is_set is True:
            self.viewer.current_mask_pixmap = self.viewer.create_mask_pixmap(
                self.viewer.maskArray[self.viewer.slice_num, :, :])
            self.viewer.update_overlay(self.viewer.current_mask_pixmap)









from piscat.InputOutput import read_status_line, reading_videos
from piscat.InputOutput import read_write_data
from piscat.BackgroundCorrection import DRA
from piscat.Preproccessing import normalization, filtering
from piscat.Localization import localization_filtering
from piscat.Localization import particle_localization
from piscat.Trajectory import particle_linking
from piscat.Trajectory import temporal_filtering

import os
import numpy as np


def protein_analysis(paths, video_names, hyperparameters, flags, name_mkdir):
    """
    This function analyses several videos based on the setting that the user defines in the ``hyperparameters`` and ``flags``.

    Parameters
    ----------
    paths: list
        List of all paths for videos.

    video_names: str
        list of all video names.

    hyperparameters: dic
        The dictionary is used to define different parameters for analysis. In the following you can see the example of this dictionary:

            | hyperparameters = {'function': 'dog', 'batch_size': 3000, 'min_V_shape_width': 1500,
                                    'search_range': 2, 'memory': 20, 'min_sigma': 1.3, 'max_sigma': 3, 'sigma_ratio': 1.1,
                                    'PSF_detection_thr': 4e-5, 'overlap': 0, 'outlier_frames_thr': 20, 'Mode_PSF_Segmentation': 'BOTH',
                                    'symmetric_PSFs_thr': 0.6, 'im_size_x': 72,
                                    'im_size_y': 72, 'image_format': '<u2'}

    flags: dic
         The dictionary is used to active/deactivate different parts in analyzing pipelines. In the following you can see the example of this dictionary:

            | flags = {'PN': True, 'outlier_frames_filter': True, 'Dense_Filter': True, 'symmetric_PSFs_Filter': True, 'FFT_flag': True}

    name_mkdir: str
        It defines the name of the folder that automatically creates next to each video to save the results of the analysis and setting history.

    Returns
    -------
    The following information will be saved
    
        * `hyperparameters`

        * `flags`

        * `Number of particles and PSFs after each steps`

        * `All extracted trajectories with 'HDF5' and 'Matlab' format.`

            * `MATLAB` saves array contains the following information for each particle:

            | [intensity_horizontal, intensity_vertical, particle_center_intensity,
                                particle_center_intensity_follow, particle_frame, particle_sigma, particle_X, particle_Y, particle_ID,
                                optional(fit_intensity, fit_x, fit_y, fit_X_sigma, fit_Y_sigma, fit_Bias, fit_intensity_error,
                                fit_x_error, fit_y_error, fit_X_sigma_error, fit_Y_sigma_error, fit_Bias_error)]

            * `HDF5` saves dictionary similar to the following structures:

            | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ..., ..., 'particle_ID': ...},
                        "#1": {}, ...}

        * `Table of PSFs information`

            * `particles`: pandas dataframe

            | Saving the clean data frame (x, y, frame, sigma, particle, ...)

    """
    PSFs_Particels_num = {'#Totall_PSFs': None,
                          '#PSFs_afterOutlierFramesFilter': None,
                          '#PSFs_afterDenseFilter': None,
                          '#PSFs_afterSymmetricPSFsFilter': None,
                          '#Totall_Particles': None,
                          '#Particles_after_V_shapeFilter': None,
                          '#Totall_frame_num_DRA': None}

    for p_, n_ in zip(paths, video_names):
        print("---" + p_ + "---")
        try:
            dr_mk = os.path.join(p_, name_mkdir)
            os.mkdir(dr_mk)
            print("Directory ", name_mkdir, " Created ")
        except FileExistsError:
            print("Directory ", name_mkdir, " already exists")

        s_dir_ = os.path.join(p_, name_mkdir)
        read_write_data.save_dic2json(data_dictionary=hyperparameters, path=s_dir_, name='hyperparameters')
        read_write_data.save_dic2json(data_dictionary=flags, path=s_dir_, name='flags')

        video = reading_videos.read_binary(file_name=p_ + '/' + n_, img_width=hyperparameters['im_size_x'], img_height=hyperparameters['im_size_y'],
                                           image_type=np.dtype(hyperparameters['image_format']))
        status_ = read_status_line.StatusLine(video)
        video, status_info = status_.find_status_line()

        if flags['PN']:
            video_pn, _ = normalization.Normalization(video=video).power_normalized()
        else:
            video_pn = video

        DRA_ = DRA.DifferentialRollingAverage(video=video_pn, batchSize=hyperparameters['batch_size'])

        RVideo_PN_ = DRA_.differential_rolling(FFT_flag=flags['FFT_flag'])

        if flags['filter_hotPixels']:
            RVideo_PN = filtering.Filters(RVideo_PN_).median(3)
        else:
            RVideo_PN = RVideo_PN_

        PSFs_Particels_num['#Totall_frame_num_DRA'] = RVideo_PN.shape[0]

        PSF_l = particle_localization.PSFsExtraction(video=RVideo_PN)
        df_PSFs = PSF_l.psf_detection(function=hyperparameters['function'],  # function='log', 'doh', 'dog'
                                      min_sigma=hyperparameters['min_sigma'], max_sigma=hyperparameters['max_sigma'],
                                      sigma_ratio=hyperparameters['sigma_ratio'],
                                      threshold=hyperparameters['PSF_detection_thr'],
                                      overlap=hyperparameters['overlap'], mode=hyperparameters['Mode_PSF_Segmentation'])

        if df_PSFs is not None:
            PSFs_Particels_num['#Totall_PSFs'] = df_PSFs.shape[0]
        else:
            PSFs_Particels_num['#Totall_PSFs'] = 0

        if PSFs_Particels_num['#Totall_PSFs'] > 1:
            s_filters = localization_filtering.SpatialFilter()

            if flags['outlier_frames_filter']:
                df_PSFs_s_filter = s_filters.outlier_frames(df_PSFs,
                                                            threshold=hyperparameters['outlier_frames_thr'])
                PSFs_Particels_num['#PSFs_afterOutlierFramesFilter'] = df_PSFs_s_filter.shape[0]
            else:
                df_PSFs_s_filter = df_PSFs

            if flags['Dense_Filter']:
                df_PSFs_s_filter = s_filters.dense_PSFs(df_PSFs_s_filter, threshold=0)
                PSFs_Particels_num['#PSFs_afterDenseFilter'] = df_PSFs_s_filter.shape[0]

            if flags['symmetric_PSFs_Filter'] and df_PSFs_s_filter.shape[0] != 0:
                df_PSF_2Dfit = PSF_l.fit_Gaussian2D_wrapper(PSF_List=df_PSFs, scale=5, internal_parallel_flag=True)
                df_PSFs_s_filter = s_filters.symmetric_PSFs(df_PSFs=df_PSF_2Dfit, threshold=hyperparameters['symmetric_PSFs_thr'])
                if df_PSFs_s_filter is not None:
                    PSFs_Particels_num['#PSFs_afterSymmetricPSFsFilter'] = df_PSFs_s_filter.shape[0]

            if df_PSFs_s_filter.shape[0] > 1 and df_PSFs_s_filter is not None:
                linking_ = particle_linking.Linking()
                df_PSFs_link = linking_.create_link(psf_position=df_PSFs_s_filter,
                                                    search_range=hyperparameters['search_range'],
                                                    memory=hyperparameters['memory'])
                PSFs_Particels_num['#Totall_Particles'] = linking_.trajectory_counter(df_PSFs_link)

                t_filters = temporal_filtering.TemporalFilter(video=RVideo_PN,
                                                              batchSize=hyperparameters['batch_size'])
                all_trajectories, df_PSFs_t_filter, his_all_particles = t_filters.v_trajectory(df_PSFs=df_PSFs_link,
                                                                            threshold=hyperparameters[
                                                                                'min_V_shape_width'])
                PSFs_Particels_num['#Particles_after_V_shapeFilter'] = linking_.trajectory_counter(df_PSFs_t_filter)

                read_write_data.save_mat(data=all_trajectories, path=s_dir_, name='all_trajectories')
                read_write_data.save_df2csv(df_PSFs_t_filter, path=s_dir_, name='position_PSFs')
                read_write_data.save_list_to_hdf5(list_data=all_trajectories, path=s_dir_, name='histData')

        read_write_data.save_dic2json(data_dictionary=PSFs_Particels_num, path=s_dir_, name='PSFs_Particels_num')



from piscat.GUI.Visualization import image_viewer
from piscat.Preproccessing import Normalization
from skimage.draw import circle_perimeter

from PySide2 import QtGui
from PySide2 import QtCore
from PySide2 import QtWidgets
from PySide2.QtCore import *

from tqdm import tqdm
from joblib import Parallel, delayed

import numpy as np
import pandas as pd


class Visulization_localization(QtWidgets.QMainWindow):

    def __init__(self, filename=None):
        super(Visulization_localization, self).__init__()
        self.different_views = {}
        self.current_frame_number = 0
        self.threadpool = QThreadPool()
        self.filename = filename

    @QtCore.Slot()
    def get_sliceNumber(self, frame_num):
        self.current_frame_number = frame_num

    def new_display(self, display_input, original_video, object=None, title=None, mask_status=False, position_list=None):

            if mask_status is False:

                self.different_views[title] = image_viewer.ImageViewer(display_input, original_video, title)
                if title != "PNG":
                    self.slice_num = self.different_views[title].viewer.slice_num
                    current_pixmap = self.different_views[title].viewer.create_pixmap(display_input[self.slice_num, :, :])
                    self.different_views[title].viewer.update_slice(current_pixmap)
                elif title == "PNG":
                    current_pixmap = self.different_views[title].viewer.create_pixmap(self.filename)
                    self.different_views[title].viewer.update_slice(current_pixmap)

            elif mask_status is True:

                self.different_views[title] = image_viewer.ImageViewer(display_input, original_video, title, mask=mask_status, list_psf=position_list)

                if title != "PNG":
                    self.slice_num = self.different_views[title].viewer.slice_num
                    current_pixmap = self.different_views[title].viewer.create_pixmap(display_input[self.slice_num, :, :])
                    self.different_views[title].viewer.update_slice(current_pixmap)
                elif title == "PNG":
                    current_pixmap = self.different_views[title].viewer.create_pixmap(self.filename)
                    self.different_views[title].viewer.update_slice(current_pixmap)

            if object is not None:
                self.different_views[title].currentFrame.connect(object.get_sliceNumber)

    def create_circle(self, x, y, r, shape):
        rr, cc = circle_perimeter(x, y, r, shape=shape)
        return rr, cc

    def create_circle_apply_oneFrame_pd(self, x_, y_, s_, f_):
        y = int(x_)
        x = int(y_)
        radius = int(np.sqrt(2) * s_)
        rr, cc = self.create_circle(x, y, radius, self.input_mask[0, :, :].shape)
        self.input_mask[f_, rr, cc] = True

    def create_circle_apply_toAll_pd(self, input_mask, position_df, parallel_flag):
        if position_df is not None:
            frames = position_df['frame'].tolist()
            x_positions = position_df['x'].tolist()
            y_positions = position_df['y'].tolist()
            sigmas = position_df['sigma'].tolist()
            # particle_labels = particle['particle'].tolist()

            if parallel_flag:
                self.input_mask = input_mask
                Parallel(n_jobs=-1, backend='threading', verbose=0)(
                    delayed(self.create_circle_apply_oneFrame_pd)(x_, y_, s_, f_) for x_, y_, s_, f_ in tqdm(zip(x_positions, y_positions, sigmas, frames)))
                input_mask = self.input_mask
            else:

                for x_, y_, s_, f_ in tqdm(zip(x_positions, y_positions, sigmas, frames)):
                    y = int(x_)
                    x = int(y_)
                    radius = int(np.sqrt(2) * s_)
                    rr, cc = self.create_circle(x, y, radius, input_mask[0, :, :].shape)
                    input_mask[f_, rr, cc] = True
            print("-------------------MaskArray is now updated.-------------------")
        else:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Do not find any particle!")
            self.msg_box.exec_()
        return input_mask

    def create_circle_apply_toAll_list(self, input_mask, position_list, flag_preview):
        if position_list is not None:
            if flag_preview is False:
                for i_ in range(len(position_list)):
                    frame_number = int(position_list[i_][0, 0])
                    for j_ in range(position_list[i_].shape[0]):
                        y = int(position_list[i_][j_, 1])
                        x = int(position_list[i_][j_, 2])
                        sigma = position_list[i_][j_, 3]
                        radius = int(np.sqrt(2)*sigma)
                        rr, cc = self.create_circle(y, x, radius, input_mask[0, :, :].shape)
                        input_mask[frame_number, rr, cc] = True
            else:
                frame_number = int(position_list[0, 0])
                for j_ in range(position_list.shape[0]):
                    y = int(position_list[j_, 1])
                    x = int(position_list[j_, 2])
                    sigma = position_list[j_, 3]
                    radius = int(np.sqrt(2) * sigma)
                    rr, cc = self.create_circle(y, x, radius, input_mask[0, :, :].shape)
                    input_mask[frame_number, rr, cc] = True
            print("-------------------MaskArray is now updated.-------------------")
        else:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Do not find any particle!")
            self.msg_box.exec_()
        return input_mask

    def call_norm_2_display(self, in_video):
        self.flag_update_display_video = True
        worker = Normalization.Normalization(in_video, flag_GUI=True, flag_image_specific=True)
        worker.signals.result.connect(self.update_display_video)
        self.threadpool.start(worker)
        while self.flag_update_display_video:
            QtGui.qApp.processEvents()

        self.flag_update_display_video = True

        return self.norm_v

    @QtCore.Slot(np.ndarray)
    def bg_correction_update(self, in_video, label, object):
            # self.call_norm_2_display(in_video)
            self.new_display(in_video, in_video, object=object, title=label, mask_status=True)

    def update_display_video(self, r_):
        self.norm_v = r_
        self.flag_update_display_video = False

    @QtCore.Slot()
    def update_localization_onMask(self, video_in, df_PSFs, title, flag_preview=False):
        try:
            self.maskArray = np.zeros_like(video_in, dtype="bool")

            if type(df_PSFs) is np.ndarray:
                circled_mask = self.create_circle_apply_toAll_list(self.maskArray, df_PSFs, flag_preview=True)
                print('list_done')

            elif type(df_PSFs) is pd.core.frame.DataFrame:
                if flag_preview:
                    circled_mask = self.create_circle_apply_toAll_pd(self.maskArray, df_PSFs, parallel_flag=False)

                else:
                    circled_mask = self.create_circle_apply_toAll_pd(self.maskArray, df_PSFs, parallel_flag=True)

            self.different_views[title].viewer.maskArray = circled_mask
            self.different_views[title].viewer.mask_is_set = True
            mask_pixmap = self.different_views[title].viewer.create_mask_pixmap(circled_mask[self.current_frame_number, :, :])
            self.different_views[title].viewer.update_overlay(mask_pixmap)
            self.different_views[title].get_in(df_PSFs)

        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Error in load data-frame!")
            self.msg_box.exec_()









from .camera_setting import CameraParameters
from .cpu_configurations import CPUConfigurations
from .image_to_video import Image2Video
from .read_status_line import StatusLine
from .read_write_data import save_mat
from .read_write_data import read_mat
from .read_write_data import save_dic_to_hdf5
from .read_write_data import save_list_to_hdf5
from .read_write_data import load_dict_from_hdf5
from .read_write_data import save_df2csv
from .read_write_data import save_dic2json
from .read_write_data import read_json2dic
from .read_write_data import download_tutorial_data
from .reading_videos import video_reader
from .reading_videos import read_binary
from .reading_videos import read_tif
from .reading_videos import read_avi
from .reading_videos import read_png
from .reading_videos import DirectoryType
from .write_video import write_binary
from .write_video import write_MP4
from .write_video import write_GIF
from .reading_videos import *



from __future__ import print_function

import pandas as pd
import json
import os
import sys


class CPUConfigurations():

    def __init__(self, n_jobs=-1, backend='multiprocessing', verbose=0, parallel_active=True, threshold_for_parallel_run=None):
        """
        This class generates a JSON file based on the parallel loop setting on the CPU that the user prefers.
        This JSON was used by other functions and methods to set hyperparameters in a parallel loop.
        For parallelization, PiSCAT used Joblib.

        | [1]. https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

        Parameters
        ----------
        n_jobs: int
            The maximum number of workers that can work at the same time.
            If -1, all CPU cores are available for use.

        backend: str
            Specify the implementation of the parallelization backend.
            The following backends are supported:

            * `“loky”`:
                It can induce some communication and Memory overhead when exchanging
                input and output data with the worker Python processes.

            * `“multiprocessing”`:
                It previous process-based backend based on multiprocessing.Pool. Less robust than loky.

            * `“threading”`:
                It is a very low-overhead backend but it suffers from the Python Global Interpreter.
                Lock if the called function relies a lot on Python objects. “threading” is mostly useful when
                the execution bottleneck is a compiled extension that explicitly releases
                the GIL (for instance a Cython loop wrapped in a “with nogil” block or an expensive call to a library such as NumPy).

        verbose: int, optional
            The verbosity level, if non zero, progress messages are printed. Above 50, the output is sent to stdout.
            The frequency of the messages increases with the verbosity level. If it more than 10, all iterations are reported.

        parallel_active: bool
            Functions will run the parallel implementation if it is True.

        threshold_for_parallel_run: float
            It reserved for next generation of PiSCAT.
        """
        try:
            self.read_cpu_setting()

        except FileNotFoundError:
            self.n_jobs = n_jobs
            self.backend = backend
            self.verbose = verbose
            self.parallel_active = parallel_active
            self.threshold_for_parallel_run = threshold_for_parallel_run

            setting_dic = {'n_jobs': [self.n_jobs], 'backend': [self.backend], 'verbose': [self.verbose],
                            'parallel_active': [self.parallel_active],
                            'threshold_for_parallel_run': [self.threshold_for_parallel_run]}

            self.save_cpu_setting(setting_dic)

    def save_cpu_setting(self, setting_dic):
        name = 'cpu_configurations.json'
        here = os.path.dirname(os.getcwd())
        subdir = 'piscat_configuration'

        try:
            dr_mk = os.path.join(here, subdir)
            os.mkdir(dr_mk)
            print("Directory ", subdir, " Created ")
        except FileExistsError:
            print("Directory ", subdir, " already exists")

        filepath = os.path.join(here, subdir, name)
        df_configfile = pd.DataFrame(data=setting_dic)
        df_configfile.to_json(filepath)

    def read_cpu_setting(self):
        subdir = "piscat_configuration"
        here = os.path.dirname(os.getcwd())
        filepath = os.path.join(here, subdir, 'cpu_configurations.json')

        with open(filepath) as json_file:
            cpu_setting = json.load(json_file)

        self.n_jobs = cpu_setting['n_jobs']['0']
        self.backend = cpu_setting['backend']['0']
        self.verbose = cpu_setting['verbose']['0']
        self.parallel_active = cpu_setting['parallel_active']['0']
        self.threshold_for_parallel_run = cpu_setting['threshold_for_parallel_run']['0']






import pickle

from piscat.BackgroundCorrection import DifferentialRollingAverage

from ..Fixtures.shared import *


def test_DifferentialRollingAverage_differential_rolling_defaults(control_video):

    # Arrange
    control_video = control_video[:2000, :, :]
    orig_video = control_video.copy()

    with open(DATA_PATH / 'differential_rolling.pck', 'rb') as f:
        expected_out_first = pickle.load(f)
        expected_out_last = pickle.load(f)
        expected_gain_first = pickle.load(f)
        expected_gain_last = pickle.load(f)

    DRA_PN = DifferentialRollingAverage(video=control_video)

    # Act
    output = DRA_PN.differential_rolling()

    # Assert

    # Test for side effects
    np.testing.assert_array_equal(orig_video, control_video)

    # Test for stable results
    np.testing.assert_array_almost_equal(output[0, ...], expected_out_first)
    np.testing.assert_array_almost_equal(output[-1, ...], expected_out_last)


def test_DifferentialRollingAverage_differential_rolling_FFT_flag(control_video):

    # Arrange
    control_video = control_video[:2000, :, :]

    expected = _load_expected(DATA_PATH / 'differential_rolling_FFT_flag.pck')

    DRA_PN = DifferentialRollingAverage(video=control_video)

    # Act
    output = DRA_PN.differential_rolling(FFT_flag=True)

    _compare(output, expected)

def _compare(output, expected):
    np.testing.assert_array_almost_equal(output[0, ...], expected[0])
    np.testing.assert_array_almost_equal(output[-1, ...], expected[1])


def _load_expected(file_name):
    with open(file_name, 'rb') as f:
        expected_out_first, expected_out_last, \
        expected_gain_first, expected_gain_last = pickle.load(f)
    return expected_out_first, expected_out_last, expected_gain_first, expected_gain_last




import pytest

import tempfile

from piscat.Tutorials.startup import *
import piscat.Tutorials.startup


def test_parse_args_working_dir(mocker):
    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script', '.'] )
        target = parse_args()
        assert Path(target).absolute() == Path.cwd()


def test_parse_args_no_working_dir(mocker):
    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script'])
        mocker.patch('piscat.Tutorials.startup.confirm', return_value=True)
        target = parse_args()
        assert Path(target).absolute() == Path.cwd()


def test_parse_args_no_working_dir_not_confirmed(mocker):
    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script'])
        mocker.patch('piscat.Tutorials.startup.confirm', return_value=False)
        with pytest.raises(Exception):
            parse_args()


def test_parse_args_dir_somewhere_else(mocker):
    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script', 'abc'])
        target = parse_args()
        assert Path(target).absolute() == Path('abc').absolute()


def test_parse_args_too_many_args_raises_exception(mocker):
    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script', 'abc', 'cba'])
        with pytest.raises(Exception):
            parse_args()


def test_find_jupyter_finds_it(mocker):
    mocker.patch('piscat.Tutorials.startup.shutil.which', return_value='jup_location')
    jupyter = find_jupyter()
    assert jupyter == 'jup_location notebook'


def test_find_jupyter_does_not_find_it(mocker):
    mocker.patch('piscat.Tutorials.startup.shutil.which', return_value=None)
    with pytest.raises(Exception):
        find_jupyter()


def test_functional_default(mocker):

    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        target = Path(tmpdir) / 'Tutorials' / 'JupyterFiles' / 'Tutorial1.ipynb'
        mocker.patch('piscat.Tutorials.startup.start_server')
        mocker.patch('piscat.Tutorials.startup.find_jupyter', return_value='/dummy/jupyter')
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script', '.'])
        assert not exists(target)
        start_tutorials()
        assert exists(target)


def test_functional_do_not_copy_when_dir_exists(mocker):

    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        os.mkdir('abc')
        target = Path(tmpdir) / 'Tutorials' / 'JupyterFiles' / 'Tutorial1.ipynb'
        mocker.patch('piscat.Tutorials.startup.start_server')
        mocker.patch('piscat.Tutorials.startup.find_jupyter', return_value='/dummy/jupyter')
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script', 'abc'])
        assert not exists(target)
        start_tutorials()
        assert not exists(target)


def test_functional_if_not_jupyter_found_no_server_called(mocker):

    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        target = Path(tmpdir) / 'Tutorials' / 'JupyterFiles' / 'Tutorial1.ipynb'
        mock_server = mocker.patch('piscat.Tutorials.startup.start_server')
        mocker.patch('piscat.Tutorials.startup.shutil.which', return_value=None)
        mocker.patch.object(piscat.Tutorials.startup.sys, 'argv', ['the_script', '.'])
        with pytest.raises(Exception):
            start_tutorials()
        assert exists(target)
        assert not mock_server.called


from __future__ import print_function
from skimage import io
from skimage.color import rgb2gray
import os
import pandas as pd
import numpy as np
import cv2


def video_reader(file_name, type='binary', img_width=128, img_height=128, image_type=np.dtype('<f8'), s_frame=0, e_frame=-1):
    """
    This is a wrapper that can be used to call various video/image readers.

    Parameters
    ----------
    file_name: str
        Path of video and file name, e.g. test.jpg.

    type: str
        Define the video/image format to be loaded.

            * 'binary': use this flag to load binary
            * 'tif': use this flag to load tif
            * 'avi': use this flag to load avi
            * 'png': use this flag to load png

    optional_parameters:
        These parameters are used when video 'bin_type' define as binary.

        img_width: int
             For binary images, it specifies the image width.

        img_height: int
            For binary images, it specifies the image height.

        image_type: str
            Numpy.dtype('<u2') --> video with uint16 pixels data type

            * "i"  (signed) integer, "u" unsigned integer, "f" floating-point
            * "<" active little-endian
            * "1" 8-bit, "2" 16-bit, "4" 32-bit, "8" 64-bit

        s_frame: int
            Video reads from this frame. This is used for cropping a video.

        e_frame: int
            Video reads until this frame. This is used for cropping a video.

    Returns
    -------
    @returns: NDArray
        The video/image
    """
    if type == 'binary':
        video = read_binary(file_name, img_width=img_width, img_height=img_height, image_type=image_type, s_frame=s_frame, e_frame=e_frame)
    elif type == 'tif':
        video = read_tif(file_name)
    elif type == 'avi':
        video = read_avi(file_name)
    elif type == 'png':
        video = read_png(file_name)
    return video


def read_binary(file_name, img_width=128, img_height=128, image_type=np.dtype('<f8'), s_frame=0, e_frame=-1):
    """
    This function reads binary video.

    Parameters
    ----------
    file_name: str
        Path and name of binary video.

    img_width: int
         It specifies the image width.

    img_height: int
        It specifies the image height.

    image_type: str

        * "i"  (signed) integer, "u" unsigned integer, "f" floating-point
        * "<" active little-endian
        * "1" 8-bit, "2" 16-bit, "4" 32-bit, "8" 64-bit

    s_frame: int
        Video reads from this frame. This is used for cropping a video.

    e_frame: int
        Video reads until this frame. This is used for cropping a video.

    Returns
    -------
    @returns: NDArray
        The video is 3D-numpy (number of frames, width, height).
    """
    if e_frame == -1:
        num_selected_frames = -1
    else:
        num_selected_frames = (e_frame - s_frame) * (img_width * img_height)
    offset = s_frame * (img_width * img_height)

    img_bin = np.fromfile(file_name, image_type, count=num_selected_frames, sep='', offset=offset)
    number_of_frame = int(img_bin.size / (img_width * img_height))
    return np.reshape(img_bin, (number_of_frame, img_width, img_height))


def read_tif(filename):
    """
    Reading image/video with TIF format.

    Parameters
    ----------
    file_name: str
        Path and name of TIF image/video.

    Returns
    -------
    @returns: NDArray
        The video is 3D-numpy (number of frames, width, height).

    """
    vid_ = io.imread(filename)
    if vid_.ndim == 4:
        frames_list = []
        for i_ in range(vid_.shape[0]):
            gray = cv2.cvtColor(vid_[i_, ...], cv2.COLOR_BGR2GRAY)
            frames_list.append(gray)
    else:
        frames_list = vid_

    frames_list = np.asarray(frames_list)
    return frames_list


def read_avi(filename):
    """
    Reading video with AVI format.

    Parameters
    ----------
    file_name: str
        Path and name of AVI video.

    Returns
    -------
    video: NDArray
        The video is 3D-numpy (number of frames, width, height).
    """

    cap = cv2.VideoCapture(filename)
    frames_list = []
    while(cap.isOpened()):
        ret, frame = cap.read()
        if ret and frame is not None:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            frames_list.append(gray)
        else:
            break

    cap.release()
    video = np.asarray(frames_list)
    return video


def read_png(filename):
    """
    Reading image with PNG format.

    Parameters
    ----------
    file_name: str
        Path and name of PNG image.

    Returns
    -------
    @returns: NDArray
        The video is 2D-numpy (width, height).
    """
    img = io.imread(filename)
    grayscale = rgb2gray(img)
    return grayscale


class DirectoryType:

    def __init__(self, dirName, type_file):
        """
        Based on the type file description, this class generates a dataframe
        containing 'Directory,' 'Folder,' and 'File' from all files below the define directory.

        Parameters
        ----------
        dirName: str
            A directory that is used to look for files of a particular type file.

        type_file: str
            The type of files that the user is looking for.
        """
        self.dirName = dirName
        self.type = type_file

        # Get the list of all files in directory tree at given path
        fileNames = []
        dirPaths = []
        dirnames = []
        for (dirpath, dirname, filenames) in os.walk(dirName):
            for file in filenames:
                if file.endswith(self.type):
                    dirPaths.append(dirpath)
                    fileNames.append(file)
                    dirnames.append(os.path.basename(dirpath))

        self.df = pd.DataFrame(list(zip(dirPaths, dirnames, fileNames)), columns=['Directory', 'Folder', 'File'])

    def return_df(self):
        """
        Based on the type file specification, this function returns a pandas data frame containing
        'Directory,"Folder,' and 'File' from all files below the define directory.

        Returns
        -------
            The data frame contains ('Directory', 'Folder', 'File')
        """
        return self.df

    def get_list_of_files(self, dirName):
        # create a list of file and sub directories
        # names in the given directory
        listOfFile = os.listdir(dirName)
        allFiles = list()
        # Iterate over all the entries
        for entry in listOfFile:
            # Create full path
            fullPath = os.path.join(dirName, entry)
            # If entry is a directory then get the list of files in this directory
            if os.path.isdir(fullPath):
                allFiles = allFiles + self.get_list_of_files(fullPath)
            else:
                allFiles.append(fullPath)

        return allFiles








from piscat.BackgroundCorrection.noise_floor import NoiseFloor
from piscat.Preproccessing import Normalization

from PySide2 import QtGui, QtCore, QtWidgets
import pandas as pd
import os


class Noise_Floor(QtWidgets.QMainWindow):

    def __init__(self, video, parent=None):
        super(Noise_Floor, self).__init__(parent)
        self.window = QtWidgets.QWidget()

        self.original_video = video

        self.min_radius = None
        self.max_radius = None
        self.step_radius = None
        self.radius_list = None
        self.mode = None
        self.file_path = None
        self.find_radius_update_tab_flag = True

        self.checkbox_power_normalization = QtWidgets.QCheckBox("Laser power normalization", self)

        self.checkbox_save = QtWidgets.QCheckBox("Saving as CSV", self)
        self.checkbox_save.toggled.connect(lambda: self.save_active())


        self.checkbox_loglog_scale = QtWidgets.QRadioButton("log scale")
        self.checkbox_loglog_scale.setChecked(True)
        self.checkbox_normal_scale = QtWidgets.QRadioButton("normal scale")

        self.ok = QtWidgets.QPushButton("Ok")
        self.ok.setAutoDefault(False)
        self.ok.clicked.connect(self.do_update)
        self.ok.setFixedWidth(100)

        self.plot = QtWidgets.QPushButton("Plot")
        self.plot.setAutoDefault(False)
        self.plot.clicked.connect(self.do_plot)
        self.plot.setFixedWidth(100)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)
        self.grid.addWidget(self.createSecondExclusiveGroup(), 1, 0)
        self.grid.addWidget(self.checkbox_power_normalization, 5, 0)
        self.grid.addWidget(self.checkbox_save, 6, 0)
        self.grid.addWidget(self.checkbox_loglog_scale, 7, 0)
        self.grid.addWidget(self.checkbox_normal_scale, 8, 0)
        self.grid.addWidget(self.ok, 9, 0)
        self.grid.addWidget(self.plot, 10, 0)

        self.setWindowTitle("Noise Floor")
        self.window.setLayout(self.grid)
        self.window.show()

    def __del__(self):
        del self
        print('Destructor called, Employee deleted.')

    def createFirstExclusiveGroup(self):
        self.groupBox_range = QtWidgets.QGroupBox("Give range for batch sizes:")
        self.groupBox_range.setCheckable(True)
        self.groupBox_range.setChecked(False)
        self.grid1 = QtWidgets.QGridLayout()
        self.add_line_edit1()
        self.groupBox_range.setLayout(self.grid1)

        return self.groupBox_range

    def createSecondExclusiveGroup(self):
        self.groupBox_list = QtWidgets.QGroupBox("Give list for batch sizes")
        self.groupBox_list.setCheckable(True)
        self.groupBox_list.setChecked(False)
        self.grid2 = QtWidgets.QGridLayout()
        self.add_line_edit2()
        self.groupBox_list .setLayout(self.grid2)

        return self.groupBox_list

    @QtCore.Slot()
    def do_update(self):

        if self.ok.clicked:
            self.get_values()
            if self.groupBox_range.isChecked() and not(self.groupBox_list.isChecked()):

                if self.min_radius != '' and self.max_radius != '' and self.step_radius != '':
                    self.min_radius = int(self.min_radius)
                    self.max_radius = int(self.max_radius)
                    self.step_radius = int(self.step_radius)
                    self.radius_list = list(range(self.min_radius, self.max_radius, self.step_radius))
                    self.mode = 'Range'
                    self.find_radius_update_tab_flag = False
                    self.run_noiseFloor()

                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Please insert radius size!")
                    self.msg_box.exec_()

            if self.groupBox_list.isChecked() and not(self.groupBox_range.isChecked()):

                if self.radius_list != '':
                    self.radius_list = eval(self.radius_list)
                    self.mode = 'List'
                    self.find_radius_update_tab_flag = False
                    self.run_noiseFloor()
                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Please insert radius size!")
                    self.msg_box.exec_()

            elif self.groupBox_list.isChecked() and self.groupBox_range.isChecked():
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Only one method should active!")
                self.msg_box.exec_()

            elif not(self.groupBox_list.isChecked()) and not(self.groupBox_range.isChecked()):
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Please select one of the methods!")
                self.msg_box.exec_()

    def run_noiseFloor(self):

        if self.checkbox_power_normalization.isChecked():
            self.original_video, _ = Normalization(video=self.original_video).power_normalized()

        result_flag = True
        n_jobs = os.cpu_count()
        inter_flag_parallel_active = True
        flag_first_except = False
        while result_flag:
            try:
                self.noise_floor_ = NoiseFloor(self.original_video, list_range=self.radius_list, n_jobs=None,
                                               inter_flag_parallel_active=inter_flag_parallel_active)
                if self.checkbox_loglog_scale.isChecked():
                    self.noise_floor_.plot_result(flag_log=True)
                else:
                    self.noise_floor_.plot_result(flag_log=False)

                if self.checkbox_save.isChecked() and self.file_path is not None:

                    noise_floor = {'batch size': self.noise_floor_.list_range, 'SNR': self.noise_floor_.mean}
                    noise_floor_df = pd.DataFrame(data=noise_floor)
                    noise_floor_df.to_csv(self.file_path)
                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Nothing Save!")
                    self.msg_box.exec_()

                result_flag = False
            except:
                inter_flag_parallel_active = False
                if flag_first_except:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Memory Error! Please off parallel CPU.")
                    self.msg_box.exec_()
                    result_flag = False
                flag_first_except = True

    def do_plot(self):
        try:
            if self.checkbox_loglog_scale.isChecked():
                self.noise_floor_.plot_result(flag_log=True)
            else:
                self.noise_floor_.plot_result(flag_log=False)
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Please update first!5")
            self.msg_box.exec_()

    def get_values(self):

        if self.groupBox_range.isChecked():
            self.min_radius = self.le1.text()
            self.max_radius = self.le2.text()
            self.step_radius = self.le3.text()

        elif self.groupBox_list.isChecked():
            self.radius_list = self.le4.text()

    def add_line_edit1(self):
        self.le1 = QtWidgets.QLineEdit()
        self.le1.setPlaceholderText('Min. batch size')
        self.le_1_label = QtWidgets.QLabel("Min batch size:")

        self.le2 = QtWidgets.QLineEdit()
        self.le2.setPlaceholderText('Max. batch size')
        self.le_2_label = QtWidgets.QLabel("Max batch size:")

        self.le3 = QtWidgets.QLineEdit()
        self.le3.setPlaceholderText('step')
        self.le_3_label = QtWidgets.QLabel("Stride between batch size:")

        self.grid1.addWidget(self.le_1_label, 2, 0)
        self.grid1.addWidget(self.le_2_label, 3, 0)
        self.grid1.addWidget(self.le_3_label, 4, 0)

        self.grid1.addWidget(self.le1, 2, 1)
        self.grid1.addWidget(self.le2, 3, 1)
        self.grid1.addWidget(self.le3, 4, 1)

    def add_line_edit2(self):
        self.le4 = QtWidgets.QLineEdit()
        self.le4.setPlaceholderText('list of batches size')
        self.le_4_label = QtWidgets.QLabel("List of all batches")

        self.grid2.addWidget(self.le_4_label, 0, 0)
        self.grid2.addWidget(self.le4, 1, 1)

    def save_active(self):
        if self.checkbox_save.isChecked():
            self.file_path = False

            self.file_path, _ = QtWidgets.QFileDialog.getSaveFileName(self, "Noise floor",
                                                                      QtCore.QDir.currentPath())

            self.file_path = self.file_path + '_noise_floor.csv'

    def closeEvent(self, event):
        event.accept()  # let the window close





from __future__ import print_function

import numpy as np


class StatusLine():

    def __init__(self, video):
        """
        A status line from an image frame is read.
        The last line of the Photonfocus camra picture is the status line.
        All data is returned in the form of a struct, as well as the cut-frame without the status line.

        Parameters
        ----------
        video: NDArray
            Numpy array with the following form should be used for video (number of frame, width, height).
        """

        self.video = video
        self.out_video = None
        self.status_line_row_flag = False
        self.status_line_column_flag = False
        self.camera_info = {}

    def find_status_line(self):
        """
        Returns
        -------
        self.out_video: NDArray
            Video without status line.

        self.camera_info: dic
            The dictionary that illustrates the obtain information in the status line.
        """

        if np.all(self.video[0, -1, :4] == [0xFF, 0x00, 0xAA, 0x55]):
            self.status_line_column_flag = True
            self.read_status_line(self.video[0])
            print("---Status line detected in column---")
            self.camera_info['status_line_position'] = 'column'
            return self.out_video, self.camera_info

        elif np.all(self.video[0, :4, -1] == [0xFF, 0x00, 0xAA, 0x55]):
            self.status_line_row_flag = True
            self.read_status_line(self.video[0])
            print("---Status line detected in row---")
            self.camera_info['status_line_position'] = 'row'
            return self.out_video, self.camera_info

        else:
            print("---Status line does not detect---")
            self.camera_info['status_line_position'] = ''
            return self.video, self.camera_info

    def dec2hex(self, n):
        return "%X" % int(n)

    def hex2dec(self, s):
        return int(s, 16)

    def decimalToBinary(self, num, len_):
        d2b_ = np.zeros(len_)
        d2b = bin(int(num)).replace("0b", "")
        tmp_ = [int(d_) for d_ in d2b]
        diff_len = abs(len_ - len(tmp_))
        d2b_[diff_len:] = tmp_

        return d2b_

    def read_status_line(self, frame):
        """
        Parameters
        ----------
        frame: NDArray
            First frame in video.
        """

        # constants for our camera
        clock_cycle = 80e6  # Camera Clock Cycle in Hz
        nr_taps = 2  # Nr of Taps
        if self.status_line_column_flag:
            cutframe = frame[:, 0:- 1]
            self.out_video = self.video[:, 1:- 1, :]

            width = frame.shape[0]
            converter = [1, 256, 256 * 256, 256 * 256 * 256]
            StatusLine = frame[-1, :]

        elif self.status_line_row_flag:
            cutframe = frame[0:- 1, :]
            self.out_video = self.video[:, :, 1:- 1]

            width = frame.shape[1]
            converter_ = [1, 256, 256 * 256, 256 * 256 * 256]
            converter = np.asarray(converter_).astype(np.int64)
            StatusLine = frame[:, -1].astype(np.int64)

        firstbits = sum(np.multiply(StatusLine[0:4], converter))
        NrAcc = firstbits / self.hex2dec('55AA00FF')

        if (NrAcc % 1) != 0:
            print('\n... It looks like there is no Status line!?')

        self.NrAcc = NrAcc
        self.Preamble = self.dec2hex(firstbits / NrAcc)  # Preamble (0x55AA00FF)

        if width >= 8:
            ImgCnt = sum(np.multiply(StatusLine[4:8], converter)) / NrAcc  # (Mean) Image Counter
            self.camera_info['ImgCnt'] = ImgCnt

        if width >= 12:
            RealTimeCnt = sum(np.multiply(StatusLine[8:12], converter)) / NrAcc  # (Mean) Real Time Counter in Microseconds
            self.camera_info['RealTimeCnt'] = RealTimeCnt

        if width >= 16:
            MissedTrgCnt = sum(np.multiply(StatusLine[12:16], converter)) / NrAcc  # (Mean) Missed Trigger Counter
            self.camera_info['MissedTrgCnt'] = MissedTrgCnt

        if width >= 20:
            self.ImgAvgVal = sum(np.multiply(StatusLine[16:20], converter)) / NrAcc  # (Mean) Image Average Value
            self.camera_info['RealTimeCnt'] = RealTimeCnt

        if width >= 24:
            IntTimeClk = sum(np.multiply(StatusLine[20:24], converter))/ NrAcc  # Integration Time in units of clock cycles (80MHz)
            IntTime = IntTimeClk / clock_cycle * nr_taps * 1000  # Integration Time in ms
            self.camera_info['IntTimeClk'] = IntTimeClk
            self.camera_info['IntTime'] = IntTime

        if width >= 28:
            BrstTrgNr = sum(np.multiply(StatusLine[24:28], converter)) / NrAcc  # (Mean) Burst Trigger Number
            self.camera_info['BrstTrgNr'] = BrstTrgNr

        if width >= 32:
            MissedBrstTrgCnt = sum(np.multiply(StatusLine[28:32], converter)) / NrAcc  # (Mean) Missed Burst Trigger Counter
            self.camera_info['MissedBrstTrgCnt'] = MissedBrstTrgCnt

        if width >= 36:
            ROIx_start = sum(np.multiply(StatusLine[32:36], converter)) / NrAcc  # Horizontal start position of ROI (Window.X)
            self.camera_info['ROIx_start'] = ROIx_start

        if width >= 40:
            ROIx_end = sum(np.multiply(StatusLine[36:40], converter)) / NrAcc  # Horizontal end position of ROI (= Window.X + Window.W - 1)
            NrCols = ROIx_end - ROIx_start + 1
            self.camera_info['ROIx_end'] = ROIx_end
            self.camera_info['NrCols'] = NrCols

        if width >= 44:
            ROIy_start = sum(np.multiply(StatusLine[40:44], converter)) / NrAcc  # Vertical start position of ROI (Window.Y). In MROI-mode this parameter is the start position of the first ROI
            self.camera_info['ROIy_start'] = ROIy_start

        if width >= 48:
            NrRows = sum(np.multiply(StatusLine[44:48], converter)) / NrAcc + 1  # Number of rows
            ROIy_end = ROIy_start + NrRows - 1
            self.camera_info['NrRows'] = NrRows
            self.camera_info['ROIy_end'] = ROIy_end

        if width >= 52:
            TrgSrc = sum(np.multiply(StatusLine[48:52], converter)) / NrAcc # Trigger Source (0=Free Running, 1=Interface Trigger, 2=I/O Trigger)
            self.camera_info['TrgSrc'] = TrgSrc

        if width >= 56:
            DigGain = 2 ** (sum(np.multiply(StatusLine[52:56], converter)) / NrAcc)  # Digital Gain (1x, 2x, 4x, 8x)
            self.camera_info['DigGain'] = DigGain

        if width >= 60:
            DigOffset = sum(np.multiply(StatusLine[56:60], converter)) / NrAcc  # Digital Offset
            self.camera_info['DigOffset'] = DigOffset

        if width >= 64:
            CamType = sum(np.multiply(StatusLine[60:64], converter)) / NrAcc  # Camera Type Code (Our Photonfocus "MV1-D1024E-160-CL-12" is 110)
            self.camera_info['CamType'] = CamType

        if width >= 68:
            SerialNr = sum(np.multiply(StatusLine[64:68], converter)) / NrAcc  # Camera Serial Number
            self.camera_info['SerialNr'] = SerialNr

        if width >= 80:
            tmp_ = list(range(3, -13, -1))
            tmp_0 = np.float_power(2, tmp_)
            tmp_1 = np.multiply(StatusLine[76:80], converter)
            tmp_2 = int(np.sum(tmp_1)/self.NrAcc)
            tmp_3 = self.decimalToBinary(tmp_2, 16)
            tmp_4 = np.multiply(tmp_3, tmp_0)
            FineGain = np.sum(tmp_4)
            self.camera_info['FineGain'] = FineGain

        if width >= 96:
            TrgLvl = self.decimalToBinary(int(np.sum(np.multiply(StatusLine[92:96], converter))) / NrAcc, 4)  # Trigger Level: signals level of the trigger input signals. Bit 0: ExSync (CC1): Bit 1: I/O Trigger; Bit 2: CC3; Bit 3: CC4
            self.camera_info['TrgLvl'] = TrgLvl




import psutil
import gc

from PySide2 import QtGui, QtCore, QtWidgets
from PySide2.QtWidgets import *
from PySide2 import QtCore, QtGui
from PySide2.QtGui import *
from PySide2.QtCore import *
from functools import partial


class VideoInMemory(QtWidgets.QMainWindow):
    display_trigger = QtCore.Signal(str)

    def __init__(self, list_available_video, parent=None):
        super().__init__(parent)
        self.list_available_video = list_available_video
        self.display_flag_original = False
        self.display_flag_DRA = False
        self.display_flag_memory = True

        self.clear_flag_original = False
        self.clear_flag_DRA = False
        self.clear_flag_memory = True

        self.window = QtWidgets.QWidget()

        self.le_1_label = QtWidgets.QLabel("Original video:")
        self.le_2_label = QtWidgets.QLabel("DRA video:")
        self.le_3_label = QtWidgets.QLabel("Memory usage:")

        # Display
        self.Display_1 = QtWidgets.QPushButton("Display")
        self.Display_1.setAutoDefault(False)
        self.Display_1.setDisabled(True)
        self.Display_1.clicked.connect(partial(self.do_update_display, mode="original_video"))
        if 'original_video' in self.list_available_video.keys():
            if self.list_available_video['original_video']:
                self.Display_1.setDisabled(False)

        self.Display_2 = QtWidgets.QPushButton("Display")
        self.Display_2.setAutoDefault(False)
        self.Display_2.setDisabled(True)
        self.Display_2.clicked.connect(partial(self.do_update_display, mode="DRA_video"))
        if 'DRA_video' in self.list_available_video.keys():
            if self.list_available_video['DRA_video']:
                self.Display_2.setDisabled(False)

        # clear
        self.Clear_1 = QtWidgets.QPushButton("Clear")
        self.Clear_1.setAutoDefault(False)
        self.Clear_1.setDisabled(True)
        self.Clear_1.clicked.connect(partial(self.do_update_clear, mode="original_video"))
        if 'original_video' in self.list_available_video.keys():
            if self.list_available_video['original_video']:
                self.Clear_1.setDisabled(False)

        self.Clear_2 = QtWidgets.QPushButton("Clear")
        self.Clear_2.setAutoDefault(False)
        self.Clear_2.setDisabled(True)
        self.Clear_2.clicked.connect(partial(self.do_update_clear, mode="DRA_video"))
        if 'DRA_video' in self.list_available_video.keys():
            if self.list_available_video['DRA_video']:
                self.Clear_2.setDisabled(False)

        self._ProgressBar = QtWidgets.QProgressBar(self)
        self._ProgressBar.setValue(0)
        self.setProgress()

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.le_1_label, 0, 0)
        # self.grid.addWidget(self.Clear_1, 0, 1)
        self.grid.addWidget(self.Display_1, 0, 2)

        self.grid.addWidget(self.le_2_label, 1, 0)
        # self.grid.addWidget(self.Clear_2, 1, 1)
        self.grid.addWidget(self.Display_2, 1, 2)

        self.grid.addWidget(self.le_3_label, 3, 0)
        self.grid.addWidget(self._ProgressBar, 3, 2)

        self.window.setWindowTitle("Display video in Memory")
        # self.setStyleSheet('QMainWindow{background-color: darkgray;}')
        self.window.setLayout(self.grid)
        self.window.show()

    def __del__(self):
        print('Destructor called, Employee deleted.')

    def get_in(self, list_available_video):
        self.list_available_video = list_available_video

        self.display_flag_original = False
        self.display_flag_DRA = False
        self.display_flag_memory = True

        self.clear_flag_original = False
        self.clear_flag_DRA = False
        self.clear_flag_memory = True

        # Display
        self.Display_1.setDisabled(True)
        if self.list_available_video['original_video']:
            self.Display_1.setDisabled(False)

        self.Display_2.setDisabled(True)
        if self.list_available_video['DRA_video']:
            self.Display_2.setDisabled(False)

        # clear
        self.Clear_1.setDisabled(True)
        if self.list_available_video['original_video']:
            self.Clear_1.setDisabled(False)

        self.Clear_2.setDisabled(True)
        if self.list_available_video['DRA_video']:
            self.Clear_2.setDisabled(False)

    # @QtCore.Slot(int)
    def setProgress(self, step=1000):
        gc.collect()
        memory = dict(psutil.virtual_memory()._asdict())
        value = memory['percent']
        # print(Memory)
        self._ProgressBar.setMaximum(step * 100)
        self._ProgressBar.setValue(step * value)
        self._ProgressBar.setFormat("%.02f %%" % value)
        self._ProgressBar.setAlignment(Qt.AlignCenter)

    def do_update_display(self, mode):
        if mode == 'original_video':
            self.display_flag_original = True
            self.display_flag_memory = False
            self.display_trigger.emit('original_video')

        elif mode == 'DRA_video':
            self.display_flag_DRA = True
            self.display_flag_memory = False
            self.display_trigger.emit('DRA_video')

    def do_update_clear(self, mode):
        if mode == 'original_video':
            self.clear_flag_original = True
            self.clear_flag_memory = False

        elif mode == 'DRA_video':
            self.clear_flag_DRA = True
            self.clear_flag_memory = False

    def closeEvent(self, **kwargs):
        QtCore.QCoreApplication.instance().quit()




from __future__ import print_function

from glob import glob
from joblib import Parallel, delayed
from piscat.InputOutput.reading_videos import video_reader
from piscat.InputOutput.cpu_configurations import CPUConfigurations
from tqdm.autonotebook import tqdm

import numpy as np
import os


class Image2Video():

    def __init__(self, path, file_format, width_size, height_size, image_type, reader_type):
        """
        This class reads images of a particular kind from a folder and concatenates them into a single NumPy array.

        Parameters
        ----------
        path: str
            The directory path that includes images.

        file_format: str
            Postfix of image names.

        width_size: int
            For binary images, it specifies the image width.

        height_size: int
            For binary images, it specifies the image height.

        image_type: str

            * "i"  (signed) integer, "u" unsigned integer, "f" floating-point.
            * "<" active little-endian.
            * "1" 8-bit, "2" 16-bit, "4" 32-bit, "8" 64-bit.

        reader_type: str
            Specify the video/image format to be loaded.

            * `'binary'`: use this flag to load binary
            * `'tif'`: use this flag to load tif
            * `'avi`': use this flag to load avi
            * `'png'`: use this flag to load png
        """
        self.cpu = CPUConfigurations()

        self.reader_type = reader_type

        self.path = os.path.join(path, file_format)
        self.width_size = width_size
        self.height_size = height_size
        self.type = image_type

        self.path_list = glob(self.path)
        self.img_bin = []
        temp = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(delayed(
            self.parallel_read_img)(x) for x in tqdm(self.path_list))
        self.video = np.asarray(temp)

    def __call__(self):
        return self.video

    def parallel_read_img(self, x):
        tmp = video_reader(file_name=x, type=self.reader_type,
                            img_width=self.width_size, img_height=self.height_size,
                            image_type=self.type, s_frame=0, e_frame=-1)
        if tmp.ndim == 3:
            return tmp[0]
        elif tmp.ndim == 2:
            return tmp
        else:
            raise ValueError('The shape {} does not correct'.format(tmp.shape))







from piscat.Localization import particle_localization, localization_filtering
from PySide2 import QtGui, QtCore, QtWidgets

import pandas as pd


class Localization_GUI(QtWidgets.QWidget):
    preview_localization = QtCore.Signal(object)
    update_localization = QtCore.Signal(object)
    output_setting_Tab_Localization = QtCore.Signal(object)
    output_number_PSFs_tracking = QtCore.Signal(object)
    update_tab_index = QtCore.Signal(int)

    def __init__(self, parent=None):
        super(Localization_GUI, self).__init__(parent)

        self.input_video = None
        self.min_radial = None
        self.max_radial = None
        self.rvt_kind = None
        self.highpass_size = None
        self.upsample = None
        self.rweights = None
        self.coarse_factor = None
        self.coarse_mode = None
        self.pad_mode = None
        self.min_sigma = None
        self.max_sigma = None
        self.sigma_ratio = None
        self.threshold = None
        self.overlap = None
        self.frame_num = 0
        self.setting_localization = {}
        self.PSFs_Particels_num = {}

        self.crappy_thr = None
        self.scale = None
        self.thr_sigma = None
        self.win_size_Fitting = None

        self.empty_value_box_flag = False
        self.flag_remove_box = True
        self.mode = None
        self.df_PSFs = None
        self.resize(300, 300)
        self.setWindowTitle('2D Localization')

        self.btn_FineLocalizationUpdate = QtWidgets.QPushButton("Update Fine Localization")
        self.btn_FineLocalizationUpdate.clicked.connect(self.do_fineLocalization)
        self.btn_FineLocalizationUpdate.setFixedWidth(150)

        self.btn_localizationUpdate = QtWidgets.QPushButton("Update Localization")
        self.btn_localizationUpdate.clicked.connect(self.do_localization)
        self.btn_localizationUpdate.setFixedWidth(150)

        self.btn_previewLocalization = QtWidgets.QPushButton('Preview', self)
        self.btn_previewLocalization.clicked.connect(self.previewLocalization)
        self.btn_previewLocalization.setIconSize(QtCore.QSize(24, 24))
        self.btn_previewLocalization.setFixedWidth(100)

        self.btn_filtering = QtWidgets.QPushButton('Filtering', self)
        self.btn_filtering.clicked.connect(self.do_Filtering)
        self.btn_filtering.setIconSize(QtCore.QSize(24, 24))
        self.btn_filtering.setMaximumSize(100, 70)

        self.mode_list = {"Bright PSF": "Bright", "Dark PSF": "Dark", "Bright & Dark PSF": "BOTH"}
        self.method_list = {"Difference of Gaussian": "dog", "Laplacian of Gaussian": "log", "Determinant of Hessian": "doh",
                            "Radial Symmetry": "frst_one_psf", 'RVT': 'RVT'}

        self.combo = QtWidgets.QComboBox(self)
        self.combo.addItem("-Select the method-")
        self.combo.addItem("Difference of Gaussian")
        self.combo.addItem("Laplacian of Gaussian")
        self.combo.addItem("Determinant of Hessian")
        self.combo.addItem("Radial Symmetry")
        self.combo.addItem("RVT")
        self.combo.currentIndexChanged.connect(self.on_select)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)
        self.grid.addWidget(self.createSecondExclusiveGroup(), 1, 0)
        self.grid.addWidget(self.createThirdExclusiveGroup(), 2, 0)
        self.grid.addWidget(self.createFourthExclusiveGroup(), 3, 0)

        self.setLayout(self.grid)

    def __del__(self):
        print('Destructor called, Employee deleted.')

    @QtCore.Slot(int)
    def get_sliceNumber(self, frame_number):
        self.frame_num = frame_number

    @QtCore.Slot()
    def update_in_data(self, input_video):
        self.input_video = input_video

    def createFirstExclusiveGroup(self):
        groupBox = QtWidgets.QGroupBox("PSF Localization")

        self.grid_Localization = QtWidgets.QGridLayout()
        self.grid_Localization.addWidget(self.combo, 0, 0)
        groupBox.setLayout(self.grid_Localization)
        return groupBox

    def createSecondExclusiveGroup(self):

        self.le_win_size = QtWidgets.QLineEdit()
        self.le_win_size.setPlaceholderText("win_size")
        self.le_win_size_label = QtWidgets.QLabel("Neighborhood size (px):")

        self.groupBox_fine_localization = QtWidgets.QGroupBox("Fine localization:")
        self.groupBox_fine_localization.setCheckable(True)
        self.groupBox_fine_localization.setChecked(False)

        self.gridFine_Localization = QtWidgets.QGridLayout()
        self.gridFine_Localization.addWidget(self.le_win_size_label, 0, 0)
        self.gridFine_Localization.addWidget(self.le_win_size, 0, 1)

        self.groupBox_fine_localization.setLayout(self.gridFine_Localization)
        return self.groupBox_fine_localization

    def createThirdExclusiveGroup(self):

        self.groupBox_update = QtWidgets.QGroupBox("Update:")

        self.grid_applyLocalization = QtWidgets.QGridLayout()
        self.grid_applyLocalization.addWidget(self.btn_previewLocalization, 0, 0)
        self.grid_applyLocalization.addWidget(self.btn_localizationUpdate, 0, 1)
        self.grid_applyLocalization.addWidget(self.btn_FineLocalizationUpdate, 0, 2)

        self.groupBox_update.setLayout(self.grid_applyLocalization)
        return self.groupBox_update

    def createFourthExclusiveGroup(self):

        self.checkbox_filter_double_PSF = QtWidgets.QCheckBox("Filter dense PSFs", self)
        self.checkbox_filter_side_lobes_PSF = QtWidgets.QCheckBox("Filter side lobes PSFs", self)
        self.checkbox_filter_asymmetry_PSF = QtWidgets.QCheckBox("Filter asymmetry PSFs", self)
        self.checkbox_2DFitting = QtWidgets.QCheckBox("2D Gaussian Fitting", self)
        self.checkbox_crappy_frames = QtWidgets.QCheckBox("Filter outlier frames", self)

        self.checkbox_filter_asymmetry_PSF.toggled.connect(lambda: self.add_line_asymmetry_PSF())
        self.checkbox_2DFitting.toggled.connect(lambda: self.add_line_2DFitting())
        self.checkbox_crappy_frames.toggled.connect(lambda: self.add_line_crappy_frames())

        self.groupBox_filters = QtWidgets.QGroupBox("Spatial filters:")
        self.groupBox_filters.setCheckable(True)
        self.groupBox_filters.setChecked(False)

        self.grid_filters = QtWidgets.QGridLayout()
        self.grid_filters.addWidget(self.checkbox_crappy_frames, 0, 0)
        self.grid_filters.addWidget(self.checkbox_filter_double_PSF, 1, 0)
        self.grid_filters.addWidget(self.checkbox_filter_side_lobes_PSF, 2, 0)
        self.grid_filters.addWidget(self.checkbox_filter_asymmetry_PSF, 3, 0)
        self.grid_filters.addWidget(self.checkbox_2DFitting, 4, 0)
        self.grid_filters.addWidget(self.btn_filtering, 5, 0)

        self.groupBox_filters.setLayout(self.grid_filters)

        return self.groupBox_filters

    def add_line_crappy_frames(self):
        if self.checkbox_crappy_frames.isChecked():
            self.line_edit_crappy = QtWidgets.QLineEdit(self)
            self.line_edit_crappy.setPlaceholderText('Max number PSFs')
            self.line_edit_crappy_label = QtWidgets.QLabel("Max number PSFs:")

            self.grid_filters.addWidget(self.line_edit_crappy_label, 0, 1)
            self.grid_filters.addWidget(self.line_edit_crappy, 0, 2)

        if not self.checkbox_crappy_frames.isChecked():

            for i_ in range(1, 5, 1):
                layout = self.grid_filters.itemAtPosition(0, i_)
                if layout is not None:
                    layout.widget().deleteLater()
                    self.grid_filters.removeItem(layout)

    def add_line_asymmetry_PSF(self):
        if self.checkbox_filter_asymmetry_PSF.isChecked():
            self.line_asymmetry_PSF = QtWidgets.QLineEdit(self)
            self.line_asymmetry_PSF.setPlaceholderText('Scale based on sigma size')
            self.line_asymmetry_PSF_label = QtWidgets.QLabel("Win_size (px):")

            self.line_asymmetry_PSF_Thr = QtWidgets.QLineEdit(self)
            tmp_str = "\u03C3"
            self.line_asymmetry_PSF_Thr.setPlaceholderText(tmp_str + "_x" + "/" + tmp_str + "_y")
            self.line_asymmetry_PSF_Thr_label = QtWidgets.QLabel("Asymmetry threshold:")

            self.grid_filters.addWidget(self.line_asymmetry_PSF_label, 3, 1)
            self.grid_filters.addWidget(self.line_asymmetry_PSF, 3, 2)

            self.grid_filters.addWidget(self.line_asymmetry_PSF_Thr_label, 3, 3)
            self.grid_filters.addWidget(self.line_asymmetry_PSF_Thr, 3, 4)

        if not self.checkbox_filter_asymmetry_PSF.isChecked():

            for i_ in range(1, 5, 1):
                layout = self.grid_filters.itemAtPosition(3, i_)
                if layout is not None:
                    layout.widget().deleteLater()
                    self.grid_filters.removeItem(layout)

    def add_line_2DFitting(self):
        if self.checkbox_2DFitting.isChecked():
            self.line_2DFitting = QtWidgets.QLineEdit(self)
            self.line_2DFitting.setPlaceholderText('Scale based on sigma size')
            self.line_2DFitting_label = QtWidgets.QLabel("Win_size (px):")

            self.grid_filters.addWidget(self.line_2DFitting_label, 4, 1)
            self.grid_filters.addWidget(self.line_2DFitting, 4, 2)

        if not self.checkbox_2DFitting.isChecked():

            for i_ in range(1, 5, 1):
                layout = self.grid_filters.itemAtPosition(4, i_)
                if layout is not None:
                    layout.widget().deleteLater()
                    self.grid_filters.removeItem(layout)

    def on_select(self):

        if self.combo.currentText() == "Difference of Gaussian":

            while self.flag_remove_box:
                self.remove_extra_box()

            self.flag_remove_box = True
            self.combo_item = self.method_list[self.combo.currentText()]
            self.PSF_mode()
            self.create_como_values()

        elif self.combo.currentText() == "Laplacian of Gaussian":

            while self.flag_remove_box:
                self.remove_extra_box()

            self.flag_remove_box = True
            self.combo_item = self.method_list[self.combo.currentText()]
            self.PSF_mode()
            self.create_como_values()

        elif self.combo.currentText() == "Determinant of Hessian":

            while self.flag_remove_box:
                self.remove_extra_box()

            self.flag_remove_box = True
            self.combo_item = self.method_list[self.combo.currentText()]
            self.create_como_values()

        elif self.combo.currentText() == "Radial Symmetry":

            while self.flag_remove_box:
                self.remove_extra_box()

            self.flag_remove_box = True
            self.combo_item = self.method_list[self.combo.currentText()]
            self.msg_box2 = QtWidgets.QMessageBox()
            self.msg_box2.setWindowTitle("Warning!")
            self.msg_box2.setText("This function only works when you have only one PSF in each frame!")
            self.msg_box2.exec_()

        elif self.combo.currentText() == "RVT":

            while self.flag_remove_box:
                self.remove_extra_box()

            self.flag_remove_box = True
            self.combo_item = self.method_list[self.combo.currentText()]
            self.create_rvt_como_values()

    def PSF_mode(self):
        self.combo_mode = QtWidgets.QComboBox(self)
        self.combo_mode.addItem("-Select the Mode-")
        self.combo_mode.addItem("Bright PSF")
        self.combo_mode.addItem("Dark PSF")
        self.combo_mode.addItem("Bright & Dark PSF")
        self.combo_mode.currentIndexChanged.connect(self.on_select_mode)
        self.grid_Localization.addWidget(self.combo_mode, 0, 1)

    def on_select_mode(self):
        self.mode = self.mode_list[self.combo_mode.currentText()]

    def create_como_values(self):
        self.le_1 = QtWidgets.QLineEdit()
        self.le_1.setPlaceholderText("Min Sigma")
        self.le_1_label = QtWidgets.QLabel("Min Sigma (px): ")

        self.le_2 = QtWidgets.QLineEdit()
        self.le_2.setPlaceholderText("Max Sigma")
        self.le_2_label = QtWidgets.QLabel("Max Sigma (px): ")

        self.le_3 = QtWidgets.QLineEdit()
        self.le_3.setPlaceholderText("Sigma Ratio/Num Sigma")
        self.le_3_label = QtWidgets.QLabel("Sigma Ratio/Num Sigma: ")

        self.le_4 = QtWidgets.QLineEdit()
        self.le_4.setPlaceholderText("Threshold")
        self.le_4_label = QtWidgets.QLabel("Threshold: ")

        self.le_5 = QtWidgets.QLineEdit()
        self.le_5.setPlaceholderText("Overlap")
        self.le_5_label = QtWidgets.QLabel("Overlap (px): ")

        self.grid_Localization.addWidget(self.le_1_label, 1, 0)
        self.grid_Localization.addWidget(self.le_2_label, 2, 0)
        self.grid_Localization.addWidget(self.le_3_label, 3, 0)
        self.grid_Localization.addWidget(self.le_4_label, 1, 2)
        self.grid_Localization.addWidget(self.le_5_label, 2, 2)

        self.grid_Localization.addWidget(self.le_1, 1, 1)
        self.grid_Localization.addWidget(self.le_2, 2, 1)
        self.grid_Localization.addWidget(self.le_3, 3, 1)
        self.grid_Localization.addWidget(self.le_4, 1, 3)
        self.grid_Localization.addWidget(self.le_5, 2, 3)

    def create_rvt_como_values(self):
        self.le_1 = QtWidgets.QLineEdit()
        self.le_1.setPlaceholderText("Min radius")
        self.le_1_label = QtWidgets.QLabel("Min radius (px): ")

        self.le_2 = QtWidgets.QLineEdit()
        self.le_2.setPlaceholderText("Max radius")
        self.le_2_label = QtWidgets.QLabel("Max radius (px): ")

        self.le_3 = QtWidgets.QLineEdit()
        self.le_3.setPlaceholderText("highpass_size")
        self.le_3_label = QtWidgets.QLabel("Highpass_size: ")

        self.le_4 = QtWidgets.QLineEdit()
        self.le_4.setPlaceholderText("upsample")
        self.le_4_label = QtWidgets.QLabel("Upsample: ")

        self.le_5 = QtWidgets.QLineEdit()
        self.le_5.setPlaceholderText("rweights")
        self.le_5_label = QtWidgets.QLabel("rweights: ")

        self.le_6 = QtWidgets.QLineEdit()
        self.le_6.setPlaceholderText("coarse_factor")
        self.le_6_label = QtWidgets.QLabel("Coarse_factor: ")

        self.le_7 = QtWidgets.QLineEdit()
        self.le_7.setPlaceholderText("Threshold")
        self.le_7_label = QtWidgets.QLabel("Threshold: ")

        self.pad_mode_group = QtWidgets.QButtonGroup()
        self.radio_pad_mode_constant = QtWidgets.QRadioButton("Pad mode (constant)")
        self.radio_pad_mode_reflect = QtWidgets.QRadioButton("Pad mode (reflect)")
        self.radio_pad_mode_edge = QtWidgets.QRadioButton("Pad mode (edge)")
        self.radio_pad_mode_fast = QtWidgets.QRadioButton("Pad mode (fast)")

        self.pad_mode_group.addButton(self.radio_pad_mode_constant)
        self.pad_mode_group.addButton(self.radio_pad_mode_reflect)
        self.pad_mode_group.addButton(self.radio_pad_mode_edge)
        self.pad_mode_group.addButton(self.radio_pad_mode_fast)
        self.radio_pad_mode_constant.setChecked(True)

        self.coarse_mode_group = QtWidgets.QButtonGroup()
        self.radio_coarse_mode_add = QtWidgets.QRadioButton("Coarse mode (add)")
        self.radio_coarse_mode_skip = QtWidgets.QRadioButton("Coarse mode (skip)")

        self.coarse_mode_group.addButton(self.radio_coarse_mode_add)
        self.coarse_mode_group.addButton(self.radio_coarse_mode_skip)
        self.radio_coarse_mode_add.setChecked(True)

        self.kind_group = QtWidgets.QButtonGroup()
        self.radio_kind_basic = QtWidgets.QRadioButton("Kind (basic)")
        self.radio_kind_normalized = QtWidgets.QRadioButton("Kind (normalized)")

        self.kind_group.addButton(self.radio_kind_basic)
        self.kind_group.addButton(self.radio_kind_normalized)
        self.radio_kind_basic.setChecked(True)

        self.grid_Localization.addWidget(self.le_1_label, 1, 0)
        self.grid_Localization.addWidget(self.le_2_label, 2, 0)
        self.grid_Localization.addWidget(self.le_3_label, 3, 0)
        self.grid_Localization.addWidget(self.le_4_label, 1, 2)
        self.grid_Localization.addWidget(self.le_5_label, 2, 2)
        self.grid_Localization.addWidget(self.le_6_label, 3, 2)
        self.grid_Localization.addWidget(self.le_7_label, 5, 2)

        self.grid_Localization.addWidget(self.le_1, 1, 1)
        self.grid_Localization.addWidget(self.le_2, 2, 1)
        self.grid_Localization.addWidget(self.le_3, 3, 1)
        self.grid_Localization.addWidget(self.le_4, 1, 3)
        self.grid_Localization.addWidget(self.le_5, 2, 3)
        self.grid_Localization.addWidget(self.le_6, 3, 3)
        self.grid_Localization.addWidget(self.le_7, 5, 3)

        self.grid_Localization.addWidget(self.radio_pad_mode_constant, 4, 0)
        self.grid_Localization.addWidget(self.radio_pad_mode_reflect, 4, 1)
        self.grid_Localization.addWidget(self.radio_pad_mode_edge, 4, 2)
        self.grid_Localization.addWidget(self.radio_pad_mode_fast, 4, 3)

        self.grid_Localization.addWidget(self.radio_coarse_mode_add, 5, 0)
        self.grid_Localization.addWidget(self.radio_coarse_mode_skip, 5, 1)

        self.grid_Localization.addWidget(self.radio_kind_basic, 6, 0)
        self.grid_Localization.addWidget(self.radio_kind_normalized, 6, 1)

    def remove_extra_box(self):
        layout_combo = self.grid_Localization.itemAtPosition(0, 1)
        if layout_combo is not None:
            layout_combo.widget().deleteLater()
            self.grid_Localization.removeItem(layout_combo)

        layout_pre_label = self.grid_Localization.itemAtPosition(1, 0)
        layout_pre = self.grid_Localization.itemAtPosition(1, 1)

        if layout_pre_label is not None:
            layout_pre_label.widget().deleteLater()
            self.grid_Localization.removeItem(layout_pre_label)

        if layout_pre is not None:
            layout_pre.widget().deleteLater()
            self.grid_Localization.removeItem(layout_pre)

        for i_ in range(1, 7, 1):
            for j_ in range(0, 6, 1):
                layout = self.grid_Localization.itemAtPosition(i_, j_)
                if layout is not None:
                    layout.widget().deleteLater()
                    self.grid_Localization.removeItem(layout)
        self.flag_remove_box = False

    def get_localization_method_parameters(self):
        try:
            if self.combo.currentText() != "Radial Symmetry" and self.combo.currentText() != "RVT":

                self.min_sigma = eval(self.le_1.text())
                self.max_sigma = eval(self.le_2.text())
                self.sigma_ratio = float(self.le_3.text())
                self.threshold = float(self.le_4.text())
                self.overlap = float(self.le_5.text())

            elif self.combo.currentText() == "RVT":
                self.min_radial = eval(self.le_1.text())
                self.max_radial = eval(self.le_2.text())

                if self.radio_kind_basic.isChecked():
                    self.rvt_kind = 'basic'
                elif self.radio_kind_normalized.isChecked():
                    self.rvt_kind = 'normalized'

                self.highpass_size = self.le_3.text()
                if self.highpass_size == '':
                    self.highpass_size = None
                else:
                    self.highpass_size = float(self.highpass_size)

                self.upsample = self.le_4.text()
                if self.upsample == '':
                    self.upsample = 1
                else:
                    self.upsample = int(self.upsample)

                self.rweights = self.le_5.text()
                if self.rweights == '':
                    self.rweights = None
                else:
                    self.rweights = eval(self.rweights)

                self.coarse_factor = self.le_6.text()
                if self.coarse_factor == '':
                    self.coarse_factor = 1
                else:
                    self.coarse_factor = float(self.coarse_factor)

                if self.radio_coarse_mode_add.isChecked():
                    self.coarse_mode = 'add'
                elif self.radio_coarse_mode_skip.isChecked():
                    self.coarse_mode = 'skip'

                if self.radio_pad_mode_constant.isChecked():
                    self.pad_mode = "constant"
                elif self.radio_pad_mode_reflect.isChecked():
                    self.pad_mode = "reflect"
                elif self.radio_pad_mode_edge.isChecked():
                    self.pad_mode = "edge"
                elif self.radio_pad_mode_fast.isChecked():
                    self.pad_mode = "fast"

                if self.le_7.text() == '':
                    self.threshold = 0
                else:
                    self.threshold = float(self.le_7.text())

            if self.groupBox_fine_localization.isChecked() and self.combo.currentText() != "Radial Symmetry":

                self.win_size = int(self.le_win_size.text())

            elif self.combo.currentText() == "Radial Symmetry" and self.groupBox_fine_localization.isChecked():

                self.win_size = int(self.le_win_size.text())

            elif self.combo.currentText() == "Radial Symmetry" and not self.groupBox_fine_localization.isChecked():

                pass

            self.empty_value_box_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all parameters!")
            self.msg_box3.exec_()

            self.empty_value_box_flag = False

    def get_values_crappy_frames(self):
        try:
            self.crappy_thr = int(self.line_edit_crappy.text())
            self.empty_value_box_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all parameters!")
            self.msg_box3.exec_()

            self.empty_value_box_flag = False

    def get_values_asymmetry_filtering(self):
        try:
            self.scale = int(self.line_asymmetry_PSF.text())
            self.thr_sigma = float(self.line_asymmetry_PSF_Thr.text())
            self.empty_value_box_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all parameters!")
            self.msg_box3.exec_()

            self.empty_value_box_flag = False

    def get_values_2DFitting(self):
        try:
            self.win_size_Fitting = int(self.line_2DFitting.text())
            self.empty_value_box_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all parameters!")
            self.msg_box3.exec_()

            self.empty_value_box_flag = False

    def removekey(self, d, key):
        r = dict(d)
        try:
            del r[key]
        except:
            print('there is no key with the name ' + key)
        return r

    @QtCore.Slot()
    def previewLocalization(self):

        layout_combo = self.grid_Localization.itemAtPosition(0, 1)
        if layout_combo is not None:
            if self.combo.currentText() == "-Select the method-" or self.combo_mode.currentText() == "-Select the Mode-":
                self.msg_box1 = QtWidgets.QMessageBox()
                self.msg_box1.setWindowTitle("Warning!")
                self.msg_box1.setText("Please select one of the methods and modes.")
                self.msg_box1.exec_()
                flag_run = False
            else:
                flag_run = True

        elif layout_combo is None:
            if self.combo.currentText() == "-Select the method-":
                self.msg_box1 = QtWidgets.QMessageBox()
                self.msg_box1.setWindowTitle("Warning!")
                self.msg_box1.setText("Please select one of the methods.")
                self.msg_box1.exec_()
                flag_run = False
            else:
                flag_run = True

        if flag_run:

            if self.input_video is None:
                self.msg_box3 = QtWidgets.QMessageBox()
                self.msg_box3.setWindowTitle("Warning!")
                self.msg_box3.setText("Please update the previous tab!")
                self.msg_box3.exec_()
            else:
                self.get_localization_method_parameters()
                if self.empty_value_box_flag:
                    psf = particle_localization.PSFsExtraction(self.input_video)
                    self.df_PSFs_preview = psf.psf_detection_preview(function=self.combo_item, min_sigma=self.min_sigma,
                                                                     max_sigma=self.max_sigma, sigma_ratio=self.sigma_ratio,
                                                                     threshold=self.threshold, overlap=self.overlap,
                                                                     frame_number=self.frame_num, mode=self.mode,
                                                                     min_radial=self.min_radial, max_radial=self.max_radial,
                                                                     highpass_size=self.highpass_size, upsample=self.upsample,
                                                                     rweights=self.rweights, coarse_factor=self.coarse_factor,
                                                                     coarse_mode=self.coarse_mode, pad_mode=self.pad_mode,
                                                                     rvt_kind=self.rvt_kind)
                    if self.groupBox_fine_localization.isChecked():
                        self.df_PSFs_preview = psf.improve_localization_with_frst(df_PSFs=self.df_PSFs_preview,
                                                                                  scale=self.win_size, flag_preview=True)

                    self.setting_localization['function'] = self.combo_item
                    self.setting_localization['mode'] = self.mode
                    self.setting_localization['min_sigma'] = self.min_sigma
                    self.setting_localization['max_sigma'] = self.max_sigma
                    self.setting_localization['sigma_ratio'] = self.sigma_ratio
                    self.setting_localization['threshold'] = self.threshold
                    self.setting_localization['overlap'] = self.overlap
                    self.setting_localization['min_radius'] = self.min_radial
                    self.setting_localization['max_radius'] = self.max_radial
                    self.setting_localization['highpass_size'] = self.highpass_size
                    self.setting_localization['upsample'] = self.upsample
                    self.setting_localization['rweights'] = self.rweights
                    self.setting_localization['coarse_factor'] = self.coarse_factor
                    self.setting_localization['pad_mode'] = self.pad_mode
                    self.setting_localization['coarse_mode'] = self.coarse_mode
                    self.setting_localization['rvt_kind'] = self.rvt_kind

                    self.preview_localization.emit(self.df_PSFs_preview)
                    self.output_setting_Tab_Localization.emit(self.setting_localization)
                    self.empty_value_box_flag = False

    @QtCore.Slot()
    def do_localization(self):
        layout_combo = self.grid_Localization.itemAtPosition(0, 1)
        if layout_combo is not None:
            if self.combo.currentText() == "-Select the method-" or self.combo_mode.currentText() == "-Select the Mode-":
                self.msg_box1 = QtWidgets.QMessageBox()
                self.msg_box1.setWindowTitle("Warning!")
                self.msg_box1.setText("Please select one of the methods and modes.")
                self.msg_box1.exec_()
                flag_run = False
            else:
                flag_run = True

        elif layout_combo is None:
            if self.combo.currentText() == "-Select the method-":
                self.msg_box1 = QtWidgets.QMessageBox()
                self.msg_box1.setWindowTitle("Warning!")
                self.msg_box1.setText("Please select one of the methods.")
                self.msg_box1.exec_()
                flag_run = False
            else:
                flag_run = True

        if flag_run:
                self.get_localization_method_parameters()
                if self.empty_value_box_flag:
                    psf = particle_localization.PSFsExtraction(video=self.input_video, flag_GUI=True)
                    self.df_PSFs = psf.psf_detection(function=self.combo_item, min_sigma=self.min_sigma,
                                                                     max_sigma=self.max_sigma,
                                                                     sigma_ratio=self.sigma_ratio,
                                                                     threshold=self.threshold, overlap=self.overlap,
                                                                     mode=self.mode,
                                                                     min_radial=self.min_radial,
                                                                     max_radial=self.max_radial,
                                                                     highpass_size=self.highpass_size,
                                                                     upsample=self.upsample,
                                                                     rweights=self.rweights,
                                                                     coarse_factor=self.coarse_factor,
                                                                     coarse_mode=self.coarse_mode,
                                                                     pad_mode=self.pad_mode,
                                                                     rvt_kind=self.rvt_kind)

                    self.PSFs_Particels_num['Total_number_PSFs'] = self.df_PSFs.shape[0]

                    self.setting_localization['function'] = self.combo_item
                    self.setting_localization['mode'] = self.mode
                    self.setting_localization['min_sigma'] = self.min_sigma
                    self.setting_localization['max_sigma'] = self.max_sigma
                    self.setting_localization['sigma_ratio'] = self.sigma_ratio
                    self.setting_localization['threshold'] = self.threshold
                    self.setting_localization['overlap'] = self.overlap
                    self.setting_localization['min_radius'] = self.min_radial
                    self.setting_localization['max_radius'] = self.max_radial
                    self.setting_localization['highpass_size'] = self.highpass_size
                    self.setting_localization['upsample'] = self.upsample
                    self.setting_localization['rweights'] = self.rweights
                    self.setting_localization['coarse_factor'] = self.coarse_factor
                    self.setting_localization['pad_mode'] = self.pad_mode
                    self.setting_localization['coarse_mode'] = self.coarse_mode
                    self.setting_localization['rvt_kind'] = self.rvt_kind

                    self.update_localization.emit(self.df_PSFs)
                    self.output_setting_Tab_Localization.emit(self.setting_localization)
                    self.output_number_PSFs_tracking.emit(self.PSFs_Particels_num)
                    self.empty_value_box_flag = False

    @QtCore.Slot()
    def do_fineLocalization(self):
        if self.groupBox_fine_localization.isChecked():
            if self.df_PSFs is not None:
                self.get_localization_method_parameters()
                if self.empty_value_box_flag:
                    psf = particle_localization.PSFsExtraction(video=self.input_video, flag_GUI=True)
                    self.df_PSFs = psf.improve_localization_with_frst(df_PSFs=self.df_PSFs, scale=self.win_size,
                                                                      flag_preview=False)
                    self.setting_localization['Flag_fine_localization'] = True
                    self.setting_localization['Fine_localization_neighborhood_size (px)'] = self.win_size

                    self.update_localization.emit(self.df_PSFs)
                    self.output_setting_Tab_Localization.emit(self.setting_localization)
                    self.output_number_PSFs_tracking.emit(self.PSFs_Particels_num)
                    self.empty_value_box_flag = False
            else:
                self.msg_box1 = QtWidgets.QMessageBox()
                self.msg_box1.setWindowTitle("Warning!")
                self.msg_box1.setText("Please update localization!")
                self.msg_box1.exec_()
        else:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Please active fine localization box!")
            self.msg_box1.exec_()

    @QtCore.Slot()
    def do_Filtering(self):

        if self.input_video is None:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please update the previous tab!")
            self.msg_box3.exec_()

        elif self.df_PSFs is None or self.df_PSFs.shape[0] <= 0:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Press the 'update' or 'loading' button to localize PSFs!")
            self.msg_box3.exec_()

        else:
            s_filters = localization_filtering.SpatialFilter()

            if self.checkbox_crappy_frames.isChecked():
                self.get_values_crappy_frames()
                if self.empty_value_box_flag:
                    self.df_PSFs_s_filter = s_filters.outlier_frames(self.df_PSFs, threshold=self.crappy_thr)
                    self.setting_localization['Flag_filter_outlier_frames'] = True
                    self.setting_localization['Outlier_frames_filter_max_number_PSFs'] = self.crappy_thr
                    self.PSFs_Particels_num['#PSFs_after_outlier_frames_filtering'] = self.df_PSFs_s_filter.shape[0]
                    self.empty_value_box_flag = False

            else:
                self.df_PSFs_s_filter = self.df_PSFs

            if self.checkbox_filter_double_PSF.isChecked():
                self.df_PSFs_s_filter = s_filters.dense_PSFs(self.df_PSFs_s_filter, threshold=0)
                self.setting_localization['Flag_filter_dense_PSFs'] = True
                self.PSFs_Particels_num['#PSFs_after_dense_PSFs_filtering'] = self.df_PSFs_s_filter.shape[0]

            if self.checkbox_filter_side_lobes_PSF.isChecked():
                self.df_PSFs_s_filter = s_filters.remove_side_lobes_artifact(self.df_PSFs_s_filter, threshold=0)
                self.setting_localization['Flag_filter_side_lobes_PSFs'] = True
                self.PSFs_Particels_num['#PSFs_after_side_lobes_PSFs_filtering'] = self.df_PSFs_s_filter.shape[0]

            if self.checkbox_2DFitting.isChecked() and self.checkbox_filter_asymmetry_PSF.isChecked():
                 pass

            if self.checkbox_2DFitting.isChecked():
                self.get_values_2DFitting()
                if self.empty_value_box_flag:
                    psf_localization = particle_localization.PSFsExtraction(video=self.input_video)
                    self.df_PSFs_s_filter = psf_localization.fit_Gaussian2D_wrapper(PSF_List=self.df_PSFs_s_filter,
                                                                                    scale=self.win_size_Fitting,
                                                                                    internal_parallel_flag=True)

                    self.setting_localization['Flag_fit_2DGaussian'] = True
                    self.setting_localization['Fit_2DGaussian_win_size (px) '] = self.scale
                    self.empty_value_box_flag = False

            if self.checkbox_filter_asymmetry_PSF.isChecked() and self.checkbox_2DFitting.isChecked():
                self.get_values_asymmetry_filtering()
                if self.empty_value_box_flag:
                    self.df_PSFs_s_filter = s_filters.symmetric_PSFs(self.df_PSFs_s_filter, threshold=self.thr_sigma)

                    self.setting_localization['Flag_filter_asymmetry_PSFs'] = True
                    self.setting_localization['Asymmetry_PSFs_filtering_threshold'] = self.thr_sigma
                    self.setting_localization['Asymmetry_PSFs_filtering_win_size (px)'] = self.scale
                    self.PSFs_Particels_num['#PSFs_after_asymmetry_PSFs_filtering'] = self.df_PSFs_s_filter.shape[0]
                    self.empty_value_box_flag = False

            self.update_localization.emit(self.df_PSFs_s_filter)
            self.output_setting_Tab_Localization.emit(self.setting_localization)
            self.output_number_PSFs_tracking.emit(self.PSFs_Particels_num)

    def closeEvent(self, event):
        QtCore.QCoreApplication.instance().quit()



from __future__ import print_function

import numpy as np
import cv2
import scipy.signal
import scipy.ndimage
import scipy.fftpack

from PySide2.QtCore import *
from tqdm.autonotebook import tqdm
from skimage import filters
from scipy.ndimage.filters import median_filter
from joblib import Parallel, delayed
from piscat.InputOutput.cpu_configurations import CPUConfigurations


class WorkerSignals(QObject):

    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)


class Filters():

    def __init__(self, video, inter_flag_parallel_active=True):
        """
       This class generates a list of video/image filters.
       To improve performance on large video files, some of them have a parallel implementation.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        inter_flag_parallel_active: bool
            If the user wants to enable general parallel tasks in the CPU configuration, he or she can only use this flag to enable or disable this process.
        """
        self.cpu = CPUConfigurations()
        self.inter_flag_parallel_active = inter_flag_parallel_active

        self.video = video
        self.filtered_video = None

    def temporal_median(self):
        """
        By extracting the temporal median from pixels, the background is corrected.

        Returns
        -------
        @returns: NDArray
            The background corrected video as 3D-numpy
        """
        video_med = np.median(self.video, axis=0)
        video_med_ = np.expand_dims(video_med, axis=0)
        return np.divide(self.video, video_med_) - 1

    def flat_field(self, sigma):
        """
        This method corrects the video background by creating a synthetic flat fielding version of the background.

        Parameters
        ----------
        sigma: float
            Sigma of Gaussian filter that use to create blur video.

        Returns
        -------
        flat_field_video: NDArray
            The background corrected video as 3D-numpy

        """
        blur_video = self.gaussian(sigma)
        flat_field_video = np.divide(self.video, blur_video) - 1
        return flat_field_video

    def median(self, size):
        """
        This function applies a 2D median filter on each frame.

        Parameters
        ----------
        size: int
            Kernel size of the median filter.

        Returns
        -------
        self.blur_video: NDArray
            The filter video as 3D-numpy.

        """
        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start median filter with Parallel---")

            result = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(delayed(
                self.median_kernel)(x, size) for x in tqdm(range(self.video.shape[0])))

            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(arry_result, (len(result), self.video.shape[1], self.video.shape[2]))
        else:
            print("\n---start median filter without Parallel---")

            result = [median_filter(self.video[i_, :, :], size) for i_ in range(self.video.shape[0])]
            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(arry_result, (len(result), self.video.shape[1], self.video.shape[2]))
        return self.filtered_video

    def median_kernel(self, i_, size):
        return median_filter(self.video[i_, :, :], size)

    def gaussian(self, sigma):
        """
        This function applies a 2D Gaussian filter on each frame.

        Parameters
        ----------
        sigma: float
            Sigma of Gaussian filter that use to create blur video.

        Returns
        -------
        self.blur_video: NDArray
            The filter video as 3D-numpy.

        """

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start gaussian filter with Parallel---")

            result = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(delayed(
                self.gaussian_kernel)(x, sigma) for x in tqdm(range(self.video.shape[0])))

            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(arry_result, (len(result), self.video.shape[1], self.video.shape[2]))
        else:
            print("\n---start gaussian filter without Parallel---")

            result = [filters.gaussian(self.video[i_, :, :], sigma=sigma) for i_ in range(self.video.shape[0])]
            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(arry_result, (len(result), self.video.shape[1], self.video.shape[2]))

        return self.filtered_video

    def gaussian_kernel(self, i_, sigma):
        return filters.gaussian(self.video[i_, :, :], sigma=sigma, preserve_range=True)


class FFT2D(QRunnable):

    def __init__(self, video):
        """
        This class computes the 2D spectrum of video.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).
        """
        super(FFT2D, self).__init__()
        self.signals = WorkerSignals()
        self.video = video

    @Slot()
    def run(self):
        result_ = self.fft2D()
        result = self.log2_scale(result_)
        self.signals.result.emit(result)

    def fft2D(self):
        im_fft = np.ndarray(self.video.shape, complex)
        for i in range(0, self.video.shape[0]):
            im_fft[i, :, :] = np.fft.fftshift(np.fft.fft2(self.video[i, :, :]))
        return im_fft

    def log2_scale(self, video):
        return np.log2(np.abs(video))


class RadialVarianceTransform():

    def __init__(self, inter_flag_parallel_active=True):
        """
        Efficient Python implementation of Radial Variance Transform.

        The main function is :func:`rvt` in the bottom of the file, which applies the transform to a single image (2D numpy array).


        Compared to the vanilla convolution implementation, there are two speed-ups:
        1) Pre-calculating and caching kernel FFT; this way so only one inverse FFT is calculated per convolution + one direct fft of the image is used for all convolutions
        2) When finding MoV, calculate ``np.mean(rsqmeans)`` in a single convolution by averaging all kernels first

        Parameters
        ----------
        inter_flag_parallel_active: bool
            In case the user wants to active general parallel tasks in CPU configuration,
            the user can only active or deactivate this method by this flag.

        References
        ----------
        [1] Kashkanova, Anna D., et al. "Precision single-particle localization using radial variance transform." Optics Express 29.7 (2021): 11070-11083.
        """
        self.cpu = CPUConfigurations()
        self._kernels_fft_cache = {}
        self.video = None
        self.inter_flag_parallel_active = inter_flag_parallel_active

    def gen_r_kernel(self, r, rmax):
        """Generate a ring kernel with radius `r` and size ``2*rmax+1``"""
        a = rmax * 2 + 1
        k = np.zeros((a, a))
        for i in range(a):
            for j in range(a):
                rij = ((i - rmax) ** 2. + (j - rmax) ** 2) ** .5
                if int(rij) == r:
                    k[i][j] = 1.

        tmp = k / np.sum(k)
        return tmp

    def generate_all_kernels(self, rmin, rmax, coarse_factor=1, coarse_mode="add"):
        """
        Generate a set of kernels with radii between `rmin` and `rmax` and sizes ``2*rmax+1``.

        ``coarse_factor`` and ``coarse_mode`` determine if the number of those kernels is reduced by either skipping or adding them
        (see :func:`rvt` for a more detail explanation).
        """
        kernels = [self.gen_r_kernel(r, rmax) for r in range(rmin, rmax + 1)]
        if coarse_factor > 1:
            if coarse_mode == "skip":
                kernels = kernels[::coarse_factor]
            else:
                kernels = [np.sum(kernels[i:i + coarse_factor], axis=0) / coarse_factor for i in
                           range(0, len(kernels), coarse_factor)]
        return kernels

    def _check_core_args(self, rmin, rmax, kind, coarse_mode="add"):
        """Check validity of the core algorithm arguments"""
        if rmin < 0 or rmax < 0:
            raise ValueError("radius should be non-negative")
        if rmin >= rmax:
            raise ValueError("rmax should be strictly greater than rmin")
        if kind not in {"basic", "normalized"}:
            raise ValueError("unrecognized kind: {}; can be either 'basic' or 'normalized'")
        if coarse_mode not in {"add", "skip"}:
            raise ValueError("unrecognized coarse mode: {}; can be either 'add' or 'skip'")

    def _check_args(self, rmin, rmax, kind, coarse_mode, highpass_size, upsample):
        """Check validity of all the algorithm arguments"""
        self._check_core_args(rmin, rmax, kind, coarse_mode)
        if upsample < 1:
            raise ValueError("upsampling factor should be positive")
        if highpass_size is not None and highpass_size < 0.3:
            raise ValueError("high-pass filter size should be >= 0.3")

    ## Prepare auxiliary parameters
    def get_fshape(self, s1, s2, fast_mode=False):
        """Get the required shape of the transformed image given the shape of the original image and the kernel"""
        shape = s1 if fast_mode else s1 + s2 - 1
        fshape = [scipy.fftpack.next_fast_len(int(d)) for d in shape]

        tmp = tuple(fshape)
        return tmp

    ## Preparing FFTs of arrays
    def prepare_fft(self, inp, fshape, pad_mode="constant"):
        """Prepare the image for a convolution by taking its Fourier transform, applying padding if necessary"""
        if pad_mode == "fast":
            return np.fft.rfftn(inp, fshape)
        else:
            pad = [((td - d) // 2, (td - d + 1) // 2) for td, d in zip(fshape, inp.shape)]
            inp = np.pad(inp, pad, mode=pad_mode)
            tmp = np.fft.rfftn(inp)
            return tmp

    ## Shortcut of SciPy fftconvolve, which takes already fft'd arrays on the input
    def convolve_fft(self, sp1, sp2, s1, s2, fshape, fast_mode=False):
        """Calculate the convolution from the Fourier transforms of the original image and the kernel, trimming the result if necessary"""
        ret = np.fft.irfftn(sp1 * sp2, fshape)
        if fast_mode:
            return np.roll(ret, (-(s2[0] // 2), -(s2[1] // 2)), (0, 1))[:s1[0], :s1[1]].copy()
        else:
            off = (fshape[0] - s1[0]) // 2 + s2[0] // 2, (fshape[1] - s1[1]) // 2 + s2[1] // 2
            tmp = ret[off[0]:off[0] + s1[0], off[1]:off[1] + s1[1]].copy()
            return tmp

    def rvt_core(self, img, rmin, rmax, kind="basic", rweights=None, coarse_factor=1, coarse_mode="add", pad_mode="constant"):
        """
        Perform core part of Radial Variance Transform (RVT) of an image.

        Parameters
        ----------
        img: NDArray
            source image (2D numpy array)

        rmin: float
            minimal radius (inclusive)

        rmax: float
            maximal radius (inclusive)

        kind: str, ("basic", "normalized")
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at lower SNR

        rweights:
            relative weights of different radial kernels; must be a 1D array of the length ``(rmax-rmin+1)//coarse_factor``

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision

        coarse_mode: str, ("add", "skip")
            The reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
            or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

        pad_mode: str, ("constant", "reflect", "edge", "fast")
            Edge padding mode for convolutions; can be either one of modes accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or ``"edge"``),
            or ``"fast"``, which means faster no-padding (a combination of ``"wrap"`` and ``"constant"`` padding depending on the image size);
            ``"fast"`` mode works faster for smaller images and larger ``rmax``, but the border pixels (within ``rmax`` from the edge) are less reliable;
            note that the image mean is subtracted before processing, so ``pad_mode="constant"`` (default) is equivalent to padding with a constant value equal to the image mean
        """
        self._check_core_args(rmin, rmax, kind, coarse_mode)  # check arguments validity
        s1 = np.array(img.shape)
        s2 = np.array([rmax * 2 + 1, rmax * 2 + 1])
        fast_mode = pad_mode == "fast"
        fshape = self.get_fshape(s1, s2,
                            fast_mode=fast_mode)  # calculate the padded image shape (add padding and get to the next "good" FFT size)
        cache_k = (rmin, rmax, coarse_factor, coarse_mode) + fshape
        if cache_k not in self._kernels_fft_cache:  # generate convolution kernels, if they are not in cache yet
            kernels = self.generate_all_kernels(rmin, rmax, coarse_factor=coarse_factor, coarse_mode=coarse_mode)
            self._kernels_fft_cache[cache_k] = [self.prepare_fft(k, fshape, pad_mode="fast") for k in
                                           kernels]  # pad_mode="fast" corresponds to the default zero-padding here
        kernels_fft = self._kernels_fft_cache[cache_k]  # get the convolution kernels (either newely generated, or cached)
        if rweights is not None:
            rweights = np.asarray(rweights)
            if len(rweights) != len(kernels_fft):
                raise ValueError(
                    "the number of kernel weights {} is different from the number of kernels {}".format(len(rweights),
                                                                                                        len(
                                                                                                            kernels_fft)))
            rweights = rweights / rweights.sum()  # normalize weights by their sum
        img = img - img.mean()  # subtract mean (makes VoM calculation more stable and zero-padding more meaningful)
        img_fft = self.prepare_fft(img, fshape, pad_mode=pad_mode)  # prepare image FFT (only needs to be done once)
        rmeans = np.array([self.convolve_fft(img_fft, k_fft, s1, s2, fshape, fast_mode=fast_mode) for k_fft in
                           kernels_fft])  # calculate M_r for all radii
        if rweights is None:
            vom = np.var(rmeans, axis=0)  # calculate VoM as a standard variance of M_r along the radius axis
        else:
            vom = np.sum(rmeans ** 2 * rweights[:, None, None], axis=0) - np.sum(rmeans * rweights[:, None, None],
                                                                                 axis=0) ** 2  # calculate VoM as a weighted variance of M_r along the radius axis
        if kind == "basic":
            return vom
        else:  # calculate MoV for normalization
            imgsq_fft = self.prepare_fft(img ** 2, fshape, pad_mode=pad_mode)  # prepare image FFT
            if rweights is None:
                sumk_fft = np.mean(kernels_fft, axis=0)  # find combined kernel as a standard mean
                mov = self.convolve_fft(imgsq_fft, sumk_fft, s1, s2, fshape, fast_mode=fast_mode) - np.mean(rmeans ** 2,
                                                                                                       axis=0)  # use the combined kernel to find MoV in one convolution
            else:
                sumk_fft = np.sum(kernels_fft * rweights[:, None, None],
                                  axis=0)  # find combined kernel as a weighted mean
                mov = self.convolve_fft(imgsq_fft, sumk_fft, s1, s2, fshape, fast_mode=fast_mode) - np.sum(
                    rmeans ** 2 * rweights[:, None, None],
                    axis=0)  # use the combined kernel to find MoV in one convolution
            tmp = vom / mov
            return tmp

    def high_pass(self, img, size):
        """Perform Gaussian high-pass filter on the image"""
        img = img.astype(float)
        tmp = img - scipy.ndimage.gaussian_filter(img, size)
        return tmp

    def rvt(self, img, rmin, rmax, kind="basic", highpass_size=None, upsample=1, rweights=None, coarse_factor=1,
            coarse_mode="add", pad_mode="constant"):
        """
        Perform Radial Variance Transform (RVT) of an image.

        Parameters
        ----------
        img: NDArray
            source image (2D numpy array)

        rmin:
            minimal radius (inclusive)

        rmax:
            maximal radius (inclusive)

        kind:
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at lower SNR

        highpass_size:
            size of the high-pass filter; ``None`` means no filter (effectively, infinite size)

        upsample: int
            integer image upsampling factor;
            `rmin` and `rmax` are adjusted automatically (i.e., they refer to the non-upsampled image);
            if ``upsample>1``, the resulting image size is multiplied by ``upsample``

        rweights:
            relative weights of different radial kernels; must be a 1D array of the length ``(rmax-rmin+1)//coarse_factor``
            coarse_factor: the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision
            coarse_mode: the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
            or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision

        coarse_mode:
            the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
            or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

        pad_mode:
            edge padding mode for convolutions; can be either one of modes accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or ``"edge"``),
            or ``"fast"``, which means faster no-padding (a combination of ``"wrap"`` and ``"constant"`` padding depending on the image size);
            ``"fast"`` mode works faster for smaller images and larger ``rmax``, but the border pixels (within ``rmax`` from the edge) are less reliable;
            note that the image mean is subtracted before processing, so ``pad_mode="constant"`` (default) is equivalent to padding with a constant value equal to the image mean

        Returns
        -------
            Returns transform source image

        """
        upsample = int(upsample)
        self._check_args(rmin, rmax, kind, coarse_mode, highpass_size, upsample)
        if highpass_size is not None:
            img = self.high_pass(img, highpass_size)
        if upsample > 1:
            img = img.repeat(upsample, axis=-2).repeat(upsample, axis=-1)  # nearest-neighbor upsampling on both axes
            if rweights is not None:
                rweights = np.asarray(rweights).repeat(upsample)  # upsample radial weights as well
            rmin *= upsample  # increase minimal and maximal radii
            rmax *= upsample

        tmp = self.rvt_core(img, rmin, rmax, kind=kind, rweights=rweights, coarse_factor=coarse_factor,
                                coarse_mode=coarse_mode, pad_mode=pad_mode)
        return tmp

    def rvt_video(self, video, rmin, rmax, kind="basic", highpass_size=None, upsample=1, rweights=None, coarse_factor=1,
                coarse_mode="add", pad_mode="constant"):
        """
        This is an RVT wrapper that allows you to get video in parallel.

        Parameters
        ----------
        video: NDArray
             The video is 3D-numpy (number of frames, width, height).

        rmin:
            minimal radius (inclusive)

        rmax:
            maximal radius (inclusive)

        kind:
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at lower SNR

        highpass_size:
            size of the high-pass filter; ``None`` means no filter (effectively, infinite size)

        upsample: int
            integer image upsampling factor;
            `rmin` and `rmax` are adjusted automatically (i.e., they refer to the non-upsampled image);
            if ``upsample>1``, the resulting image size is multiplied by ``upsample``

        rweights:
            relative weights of different radial kernels; must be a 1D array of the length ``(rmax-rmin+1)//coarse_factor``
            coarse_factor: the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision
            coarse_mode: the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
            or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision

        coarse_mode:
            the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
            or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

        pad_mode:
            edge padding mode for convolutions; can be either one of modes accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or ``"edge"``),
            or ``"fast"``, which means faster no-padding (a combination of ``"wrap"`` and ``"constant"`` padding depending on the image size);
            ``"fast"`` mode works faster for smaller images and larger ``rmax``, but the border pixels (within ``rmax`` from the edge) are less reliable;
            note that the image mean is subtracted before processing, so ``pad_mode="constant"`` (default) is equivalent to padding with a constant value equal to the image mean

        Returns
        -------
            Returns transform source video.
        """
        self.video = video

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start RVT with Parallel---")

            result = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(delayed(self.rvt_kernel)(f_,
                                                                                                                        rmin=rmin,
                                                                                                                        rmax=rmax,
                                                                                                                        kind=kind,
                                                                                                                        highpass_size=highpass_size,
                                                                                                                        upsample=upsample,
                                                                                                                        rweights=rweights,
                                                                                                                        coarse_factor=coarse_factor,
                                                                                                                        coarse_mode=coarse_mode,
                                                                                                                        pad_mode=pad_mode) for f_ in tqdm(range(video.shape[0])))
            arry_result = np.asarray(result)
            self.rvt_video = np.reshape(arry_result, (len(result), self.video.shape[1], self.video.shape[2]))
        else:
            print("\n---start RVT without Parallel---")

            result = [self.rvt_kernel(i_, rmin=rmin, rmax=rmax, kind=kind, highpass_size=highpass_size,
                                               upsample=upsample, rweights=rweights, coarse_factor=coarse_factor,
                                            coarse_mode=coarse_mode, pad_mode=pad_mode) for i_ in tqdm(range(self.video.shape[0]))]
            arry_result = np.asarray(result)
            self.rvt_video = np.reshape(arry_result, (len(result), self.video.shape[1], self.video.shape[2]))
        return self.rvt_video

    def rvt_kernel(self, frame_num, rmin, rmax, kind="basic", highpass_size=None, upsample=1, rweights=None, coarse_factor=1,
            coarse_mode="add", pad_mode="constant"):
        """
        Perform Radial Variance Transform (RVT) of an image.

        Parameters
        ----------
        frame_num: int
            frame number

        rmin:
            minimal radius (inclusive)

        rmax:
            maximal radius (inclusive)

        kind:
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at lower SNR

        highpass_size:
            size of the high-pass filter; ``None`` means no filter (effectively, infinite size)

        upsample: int
            integer image upsampling factor;
            `rmin` and `rmax` are adjusted automatically (i.e., they refer to the non-upsampled image);
            if ``upsample>1``, the resulting image size is multiplied by ``upsample``

        rweights:
            relative weights of different radial kernels; must be a 1D array of the length ``(rmax-rmin+1)//coarse_factor``
            coarse_factor: the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision
            coarse_mode: the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
            or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision

        coarse_mode:
            the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
            or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

        pad_mode:
            edge padding mode for convolutions; can be either one of modes accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or ``"edge"``),
            or ``"fast"``, which means faster no-padding (a combination of ``"wrap"`` and ``"constant"`` padding depending on the image size);
            ``"fast"`` mode works faster for smaller images and larger ``rmax``, but the border pixels (within ``rmax`` from the edge) are less reliable;
            note that the image mean is subtracted before processing, so ``pad_mode="constant"`` (default) is equivalent to padding with a constant value equal to the image mean

        Returns
        -------
            Returns transform source image

        """
        img = self.video[frame_num, :, :]
        upsample = int(upsample)
        self._check_args(rmin, rmax, kind, coarse_mode, highpass_size, upsample)
        if highpass_size is not None:
            img = self.high_pass(img, highpass_size)
        if upsample > 1:
            img = img.repeat(upsample, axis=-2).repeat(upsample, axis=-1)  # nearest-neighbor upsampling on both axes
            if rweights is not None:
                rweights = np.asarray(rweights).repeat(upsample)  # upsample radial weights as well
            rmin *= upsample  # increase minimal and maximal radii
            rmax *= upsample
        return self.rvt_core(img, rmin, rmax, kind=kind, rweights=rweights, coarse_factor=coarse_factor,
                             coarse_mode=coarse_mode, pad_mode=pad_mode)


class FastRadialSymmetryTransform():
    def __init__(self):
        '''
        Implementation of fast radial symmetry transform in pure Python using OpenCV and numpy.

        References
        ----------
        [1] Loy, G., & Zelinsky, A. (2002). A fast radial symmetry transform for detecting points of interest. Computer Vision, ECCV 2002.
        [2] https://github.com/Xonxt/frst
        '''
        pass

    def gradx(self, img):

      rows, cols = img.shape
      return np.hstack((np.zeros((rows, 1)), (img[:, 2:] - img[:, :-2])/2.0, np.zeros((rows, 1))))


    def grady(self, img):
        # img = img.astype('int')
        rows, cols = img.shape
        # Use vstack to add back the rows that were dropped as zeros
        return np.vstack( (np.zeros((1, cols)), (img[2:, :] - img[:-2, :])/2.0, np.zeros((1, cols))) )


    def _frst(self, img, radii, alpha, beta, stdFactor, mode='BOTH'):
        """
        Performs fast radial symmetry transform

        Parameters
        ----------
        img: NDArray
           Input_video image, grayscale.

        radii: int
           Integer value for radius size in pixels (n in the original paper); also used to size gaussian kernel

        alpha: float
           Strictness of symmetry transform (higher=more strict; 2 is good place to start)

        beta: float
            Gradient threshold parameter, float in [0,1]

        stdFactor: float
           Standard deviation factor for gaussian kernel

        mode: str
           BRIGHT, DARK, or BOTH
        """
        mode = mode.upper()
        assert mode in ['BRIGHT', 'DARK', 'BOTH']
        dark = (mode == 'DARK' or mode == 'BOTH')
        bright = (mode == 'BRIGHT' or mode == 'BOTH')

        workingDims = tuple((e + 2*radii) for e in img.shape)

        output = np.zeros(img.shape, np.float64)
        O_n = np.zeros(workingDims, np.float64)
        M_n = np.zeros(workingDims, np.float64)

        #Calculate gradients
        gx = self.gradx(img)
        gy = self.grady(img)

        #Find gradient vector magnitude
        gnorms = np.sqrt(np.add(np.multiply(gx, gx), np.multiply(gy, gy)))

        #Use beta to set threshold - speeds up transform significantly
        gthresh = np.amax(gnorms) * beta

        #Find x/y distance to affected pixels
        gpx = np.multiply(np.divide(gx, gnorms, out=np.zeros(gx.shape), where=gnorms!=0), radii).round().astype(int)
        gpy = np.multiply(np.divide(gy, gnorms, out=np.zeros(gy.shape), where=gnorms!=0), radii).round().astype(int)

        #Iterate over all pixels (w/ gradient above threshold)
        for coords, gnorm in np.ndenumerate(gnorms):
            if gnorm > gthresh:
              i, j = coords
              #Positively affected pixel
              if bright:
                ppve = (i+gpx[i, j], j+gpy[i, j])
                O_n[ppve] += 1
                M_n[ppve] += gnorm
              #Negatively affected pixel
              if dark:
                pnve = (i-gpx[i, j], j-gpy[i, j])
                O_n[pnve] -= 1
                M_n[pnve] -= gnorm

        #Abs and normalize O matrix
        O_n = np.abs(O_n)
        O_n = O_n / float(np.amax(O_n))

        #Normalize M matrix
        M_max = float(np.amax(np.abs(M_n)))
        M_n = M_n / M_max

        #Elementwise multiplication
        F_n = np.multiply(np.power(O_n, alpha), M_n)

        #Gaussian blur
        kSize = int(np.ceil(radii / 2))
        kSize = kSize + 1 if kSize % 2 == 0 else kSize

        S = cv2.GaussianBlur(F_n, (kSize, kSize), int(radii * stdFactor ))
        S = np.fliplr(S)

        return S[radii:-radii, radii:-radii]




